# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

# from .mpi import MPITestCase
from toast.tests.mpi import MPITestCase

import glob
import os
import shutil

import healpy as hp
import numpy as np

from ..timing import gather_timers, GlobalTimers
from ..timing import dump as dump_timing
from ..tod import AnalyticNoise, OpSimNoise, Interval, OpCacheCopy, OpCacheInit
from ..map import DistPixels
from ..todmap import (
    TODHpixSpiral,
    OpSimGradient,
    OpPointingHpix,
    OpMadam,
    OpMapMaker,
    OpSimScan,
)
from ..todmap.mapmaker import TemplateMatrix, OffsetTemplate, Signal
from .. import qarray as qa

from ._helpers import create_outdir, create_distdata, boresight_focalplane


class OpMapMakerTest(MPITestCase):
    def setUp(self):
        fixture_name = os.path.splitext(os.path.basename(__file__))[0]
        self.outdir = create_outdir(self.comm, fixture_name)
        self.rank = 0
        if self.comm is not None:
            self.rank = self.comm.rank
        if self.rank == 0:
            for fname in glob.glob("{}/*".format(self.outdir)):
                try:
                    os.remove(fname)
                except OSError:
                    pass
        if self.comm is not None:
            self.comm.barrier()

        self.nobs = 1
        self.data = create_distdata(self.comm, obs_per_group=self.nobs)

        self.ndet = self.data.comm.group_size
        self.sigma = 1
        self.rate = 50.0
        self.net = self.sigma / np.sqrt(self.rate)
        self.alpha = 2
        self.fknee = 1e0

        # Create detectors
        (
            dnames,
            dquat,
            depsilon,
            drate,
            dnet,
            dfmin,
            dfknee,
            dalpha,
        ) = boresight_focalplane(
            self.ndet,
            samplerate=self.rate,
            fknee=self.fknee,
            alpha=self.alpha,
            net=self.net,
        )

        # Pixelization
        self.sim_nside = 32
        self.map_nside = 32
        self.pointingmode = "IQU"
        self.nnz = 3

        # Samples per observation
        self.npix = 12 * self.sim_nside ** 2
        self.ninterval = 4
        self.totsamp = self.ninterval * self.npix

        # Define intervals
        intervals = []
        interval_length = self.npix
        istart = 0
        while istart < self.totsamp:
            istop = istart + interval_length
            intervals.append(
                Interval(
                    start=istart / self.rate,
                    stop=istop / self.rate,
                    first=istart,
                    last=istop - 1,
                )
            )
            istart = istop

        # Construct an uncorrelated analytic noise model for the detectors

        noise = AnalyticNoise(
            rate=drate,
            fmin=dfmin,
            detectors=dnames,
            fknee=dfknee,
            alpha=dalpha,
            NET=dnet,
        )

        # Populate the observations

        for iobs in range(self.nobs):
            if iobs % 3 == 1:
                rot = qa.from_angles(np.pi / 2, 0, 0)
            if iobs % 3 == 2:
                rot = qa.from_angles(np.pi / 2, np.pi / 2, 0)
            else:
                rot = None

            tod = TODHpixSpiral(
                self.data.comm.comm_group,
                dquat,
                self.totsamp,
                detranks=self.data.comm.group_size,
                rate=self.rate,
                nside=self.sim_nside,
                rot=rot,
            )

            self.data.obs[iobs]["tod"] = tod
            self.data.obs[iobs]["noise"] = noise
            self.data.obs[iobs]["intervals"] = intervals

        # Write processing masks

        self.npix = 12 * self.map_nside ** 2
        pix = np.arange(self.npix)
        pix = hp.reorder(pix, r2n=True)

        self.binary_mask = np.logical_or(pix < self.npix * 0.45, pix > self.npix * 0.55)
        self.maskfile_binary = os.path.join(self.outdir, "binary_mask.fits")
        hp.write_map(
            self.maskfile_binary,
            self.binary_mask,
            dtype=np.float32,
            overwrite=True,
            nest=True,
        )

        self.smooth_mask = (np.abs(pix - self.npix / 2) / (self.npix / 2)) ** 0.5
        self.maskfile_smooth = os.path.join(self.outdir, "apodized_mask.fits")
        hp.write_map(
            self.maskfile_smooth,
            self.smooth_mask,
            dtype=np.float32,
            overwrite=True,
            nest=True,
        )

        # Synthesize an input map
        self.lmax = 2 * self.sim_nside
        self.cl = np.ones([4, self.lmax + 1])
        self.cl[:, 0:2] = 0
        fwhm = np.radians(10)
        self.inmap = hp.synfast(
            self.cl,
            self.sim_nside,
            lmax=self.lmax,
            mmax=self.lmax,
            pol=True,
            pixwin=True,
            fwhm=np.radians(30),
            verbose=False,
        )
        self.inmap = hp.reorder(self.inmap, r2n=True)
        self.inmapfile = os.path.join(self.outdir, "input_map.fits")
        hp.write_map(self.inmapfile, self.inmap, overwrite=True, nest=True)

        return

    """
    def test_project_signal_offsets(self):
        from .._libtoast import project_signal_offsets

        x = np.arange(1000, dtype=np.float64)
        todslice = slice(100, 200)
        amplitudes = np.zeros(100, dtype=np.float64)
        itemplate = 10
        testvalue = np.sum(x[todslice])
        project_signal_offsets(x, [todslice], amplitudes, [itemplate])
        print("testvalue =", testvalue, ", value = ", amplitudes[itemplate], flush=True)
        # Timing
        from ..mpi import MPI

        n = 1000000
        t1 = MPI.Wtime()
        for i in range(n):
            testvalue = np.sum(x[todslice])
        t2 = MPI.Wtime()
        todslices = []
        itemplates = []
        for i in range(n):
            todslices.append(todslice)
            itemplates.append(itemplate)
        project_signal_offsets(x, todslices, amplitudes, np.array(itemplates))
        t3 = MPI.Wtime()
        print("Time1 =", t2 - t1, ", Time2 =", t3 - t2, flush=True)
    """

    """
    def test_subharmonic_template(self):

        name = "testtod1"
        init = OpCacheInit(name=name, init_val=0)

        # make a simple pointing matrix
        pointing = OpPointingHpix(
            nside=self.map_nside, nest=True, mode=self.pointingmode
        )
        pointing.exec(self.data)

        # Scan the signal from a map
        distmap = DistPixels(
            comm=self.comm,
            size=self.npix,
            nnz=self.nnz,
            dtype=np.float32,
            submap=subnpix,
            local=localsm,
        )
        distmap.read_healpix_fits(self.inmapfile)
        scansim = OpSimScan(input_map=distmap, out=name)
        scansim.exec(self.data)

        # add a sharp gradient to one of the detectors
        for obs in self.data.obs:
            tod = obs["tod"]
            sig = tod.local_signal("d01", name)
            sig[:] += np.arange(sig.size)

        # Run TOAST mapmaker

        mapmaker = OpMapMaker(
            nside=self.map_nside,
            nnz=self.nnz,
            name=name,
            outdir=self.outdir,
            outprefix="toast_subharmonic_test_",
            baseline_length=None,
            # maskfile=self.maskfile_binary,
            # weightmapfile=self.maskfile_smooth,
            subharmonic_order=1,
        )
        mapmaker.exec(self.data)

        # Run the mapmaker again

        mapmaker = OpMapMaker(
            nside=self.map_nside,
            nnz=self.nnz,
            name=name,
            outdir=self.outdir,
            outprefix="toast_subharmonic_test2_",
            baseline_length=None,
            # maskfile=self.maskfile_binary,
            # weightmapfile=self.maskfile_smooth,
            subharmonic_order=1,
        )
        mapmaker.exec(self.data)

        # Compare

        m1 = None
        if self.rank == 0:
            m1 = hp.read_map(
                os.path.join(self.outdir, "toast_subharmonic_test2_destriped.fits"),
                None,
                nest=True,
            )

        failed = False
        if self.rank == 0:
            if not np.allclose(self.inmap[1], m1[1], atol=1e-6, rtol=1e-6):
                print("Input and output maps do not agree.")
                failed = True
        if self.comm is not None:
            failed = self.comm.bcast(failed, root=0)
        self.assertFalse(failed)

        return
    """

    """
    def test_offset_template(self):

        name = "testtod2"
        init = OpCacheInit(name=name, init_val=0)

        # Add simulated noise
        opnoise = OpSimNoise(realization=0, out=name)
        opnoise.exec(self.data)

        # Build detector weights
        all_detweights = []
        for obs in self.data.obs:
            detweights = {}
            tod = obs["tod"]
            for det in tod.local_dets:
                detweights[det] = 1
            all_detweights.append(detweights)

        # Create offset template
        step_time = 1
        step_size = int(step_time * self.rate)
        offset_template = OffsetTemplate(
            self.data,
            all_detweights,
            step_length=step_time,
            intervals="intervals",
            use_noise_prior=True,
        )

        # Create Signal
        signal = Signal(self.data, name=name)

        # Create template matrix
        templates = TemplateMatrix(self.data, self.comm, [offset_template])

        amplitudes1 = templates.zero_amplitudes()
        reference = templates.apply_transpose(signal)
        reference /= step_size
        amplitudes3 = templates.zero_amplitudes()
        templates.add_prior(reference, amplitudes3)
        print("amplitudes1:", amplitudes1)
        print("reference:", reference)
        print("amplitudes3:", amplitudes3)

        return

    """

    def test_mapmaker_madam(self):

        name = "testtod2"
        init = OpCacheInit(name=name, init_val=0)

        # make a simple pointing matrix
        pointing = OpPointingHpix(
            nside=self.map_nside, nest=True, mode=self.pointingmode
        )
        pointing.exec(self.data)

        # Scan the signal from a map
        distmap = DistPixels(self.data, nnz=self.nnz, dtype=np.float32)
        distmap.read_healpix_fits(self.inmapfile)

        scansim = OpSimScan(input_map=distmap, out=name)
        scansim.exec(self.data)

        # Add simulated noise
        opnoise = OpSimNoise(realization=0, out=name)
        opnoise.exec(self.data)

        # Copy the signal to run with Madam
        name_madam = name + "_copy"
        cachecopy = OpCacheCopy(name, name_madam)
        cachecopy.exec(self.data)

        # Run TOAST mapmaker

        gt = GlobalTimers.get()
        gt.start("OpMapMaker test")

        mapmaker = OpMapMaker(
            nside=self.map_nside,
            nnz=self.nnz,
            name=name,
            outdir=self.outdir,
            outprefix="toast_test_",
            baseline_length=1,
            maskfile=self.maskfile_binary,
            # weightmapfile=self.maskfile_smooth,
            subharmonic_order=None,
            iter_max=100,
            use_noise_prior=True,
            precond_width=30,
        )
        mapmaker.exec(self.data)
        # User needs to set TOAST_FUNCTIME to see timing results
        mapmaker.report_timing()

        gt.stop_all()
        alltimers = gather_timers(comm=self.comm)
        if self.rank == 0:
            out = os.path.join(self.outdir, "timing")
            dump_timing(alltimers, out)
            print("Saved timers in {}".format(out))

        # Run the destriper again
        # mapmaker = OpMapMaker(
        #    nside=self.map_nside,
        #    nnz=self.nnz,
        #    name=name,
        #    outdir=self.outdir,
        #    outprefix="toast_test2_",
        #    baseline_length=1,
        #    maskfile=self.maskfile_binary,
        #    # weightmapfile=self.maskfile_smooth,
        #    subharmonic_order=None,
        #    iter_max=100,
        #    use_noise_prior=True,
        # )
        # mapmaker.exec(self.data)

        # Run Madam

        pars = {}
        pars["kfirst"] = "T"
        pars["base_first"] = 1.0
        pars["fsample"] = self.rate
        pars["nside_map"] = self.map_nside
        pars["nside_cross"] = self.map_nside
        pars["nside_submap"] = min(8, self.map_nside)
        pars["write_map"] = "T"
        pars["write_binmap"] = "T"
        pars["write_matrix"] = "F"
        pars["write_wcov"] = "T"
        pars["write_hits"] = "T"
        pars["file_inmask"] = self.maskfile_binary
        pars["kfilter"] = "T"
        pars["path_output"] = self.outdir
        pars["file_root"] = "madam"
        pars["info"] = 3

        madam = OpMadam(
            params=pars,
            name=name_madam,
            name_out=name_madam,
            flag_mask=1,
            detweights=mapmaker.detweights[0],
        )
        if not madam.available:
            print("libmadam not available, skipping mapmaker comparison")
            return

        madam.exec(self.data)

        # DEBUG begin
        # import pdb
        # import matplotlib.pyplot as plt
        # tod = self.data.obs[0]["tod"]
        # sig1 = tod.local_signal("d00", name)
        # sig2 = tod.local_signal("d00", name_madam)
        # plt.plot(sig1, '.', label="TOAST mapmaker")
        # plt.plot(sig2, '.', label="Madam")
        # plt.plot(sig2 - sig1, '.', label="Madam - TOAST")
        # plt.legend(loc="best")
        # plt.savefig("test2.png")
        # pdb.set_trace()
        # DEBUG end

        # Compare

        m0 = None
        if self.rank == 0:
            m0 = hp.read_map(os.path.join(self.outdir, "toast_test_binned.fits"))
        m1 = None
        if self.rank == 0:
            m1 = hp.read_map(os.path.join(self.outdir, "madam_bmap.fits"))

        failed = False
        if self.rank == 0:
            if not np.allclose(m0[self.binary_mask], m1[self.binary_mask]):
                print("TOAST mapmaker and Madam do not agree.")
                failed = True
        if self.comm is not None:
            failed = self.comm.bcast(failed, root=0)
        self.assertFalse(failed)

        return



# Copyright (c) 2019-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

import argparse
import os

import numpy as np
from scipy.constants import h, k

from ..timing import function_timer, Timer
from ..utils import Logger, Environment

from .. import qarray as qa

from ..todmap import OpSimAtmosphere, atm_available_utils

if atm_available_utils:
    from ..todmap.atm import (
        atm_atmospheric_loading,
        atm_atmospheric_loading_vec,
        atm_absorption_coefficient,
        atm_absorption_coefficient_vec,
    )


XAXIS, YAXIS, ZAXIS = np.eye(3)
TCMB = 2.725


def add_atmosphere_args(parser):
    """Add the atmospheric simulation arguments"""
    parser.add_argument(
        "--atmosphere",
        required=False,
        action="store_true",
        help="Add simulated atmoshere",
        dest="simulate_atmosphere",
    )
    parser.add_argument(
        "--simulate-atmosphere",
        required=False,
        action="store_true",
        help="Add simulated atmoshere",
        dest="simulate_atmosphere",
    )
    parser.add_argument(
        "--no-atmosphere",
        required=False,
        action="store_false",
        help="Do not add simulated atmosphere",
        dest="simulate_atmosphere",
    )
    parser.add_argument(
        "--simulate-coarse-atmosphere",
        required=False,
        action="store_true",
        help="Add simulated coarse atmoshere",
        dest="simulate_coarse_atmosphere",
    )
    parser.add_argument(
        "--no-coarse-atmosphere",
        required=False,
        action="store_false",
        help="Do not add simulated coarse atmosphere",
        dest="simulate_coarse_atmosphere",
    )
    parser.add_argument(
        "--no-simulate-atmosphere",
        required=False,
        action="store_false",
        help="Do not add simulated atmosphere",
        dest="simulate_atmosphere",
    )
    parser.set_defaults(simulate_atmosphere=False)

    parser.add_argument(
        "--focalplane-radius-deg",
        required=False,
        type=np.float,
        help="Override focal plane radius [deg]",
    )

    parser.add_argument(
        "--atm-verbosity",
        required=False,
        default=0,
        type=np.int,
        help="Atmospheric sim verbosity level",
    )
    parser.add_argument(
        "--atm-lmin-center",
        required=False,
        default=0.01,
        type=np.float,
        help="Kolmogorov turbulence dissipation scale center",
    )
    parser.add_argument(
        "--atm-lmin-sigma",
        required=False,
        default=0.001,
        type=np.float,
        help="Kolmogorov turbulence dissipation scale sigma",
    )
    parser.add_argument(
        "--atm-lmax-center",
        required=False,
        default=10.0,
        type=np.float,
        help="Kolmogorov turbulence injection scale center",
    )
    parser.add_argument(
        "--atm-lmax-sigma",
        required=False,
        default=10.0,
        type=np.float,
        help="Kolmogorov turbulence injection scale sigma",
    )
    parser.add_argument(
        "--atm-gain",
        required=False,
        default=2e-5,
        type=np.float,
        help="Atmospheric gain factor.",
    )
    parser.add_argument(
        "--atm-gain-coarse",
        required=False,
        default=8e-5,
        type=np.float,
        help="Coarse atmospheric gain factor.",
    )
    parser.add_argument(
        "--atm-zatm",
        required=False,
        default=40000.0,
        type=np.float,
        help="atmosphere extent for temperature profile",
    )
    parser.add_argument(
        "--atm-zmax",
        required=False,
        default=200.0,
        type=np.float,
        help="atmosphere extent for water vapor integration",
    )
    parser.add_argument(
        "--atm-xstep",
        required=False,
        default=10.0,
        type=np.float,
        help="size of volume elements in X direction",
    )
    parser.add_argument(
        "--atm-ystep",
        required=False,
        default=10.0,
        type=np.float,
        help="size of volume elements in Y direction",
    )
    parser.add_argument(
        "--atm-zstep",
        required=False,
        default=10.0,
        type=np.float,
        help="size of volume elements in Z direction",
    )
    parser.add_argument(
        "--atm-nelem-sim-max",
        required=False,
        default=10000,
        type=np.int,
        help="controls the size of the simulation slices",
    )
    parser.add_argument(
        "--atm-wind-dist",
        required=False,
        default=3000.0,
        type=np.float,
        help="Maximum wind drift to simulate without discontinuity",
    )
    parser.add_argument(
        "--atm-z0-center",
        required=False,
        default=2000.0,
        type=np.float,
        help="central value of the water vapor distribution",
    )
    parser.add_argument(
        "--atm-z0-sigma",
        required=False,
        default=0.0,
        type=np.float,
        help="sigma of the water vapor distribution",
    )
    parser.add_argument(
        "--atm-T0-center",
        required=False,
        default=280.0,
        type=np.float,
        help="central value of the temperature distribution",
    )
    parser.add_argument(
        "--atm-T0-sigma",
        required=False,
        default=10.0,
        type=np.float,
        help="sigma of the temperature distribution",
    )
    parser.add_argument(
        "--atm-cache",
        required=False,
        default="atm_cache",
        help="Atmosphere cache directory",
    )
    parser.add_argument(
        "--atm-apply-flags",
        default=False,
        action="store_true",
        help="Only simulate unflagged samples.",
    )
    # Common flag mask may already be added
    try:
        parser.add_argument(
            "--common-flag-mask",
            required=False,
            default=1,
            type=np.uint8,
            help="Common flag mask",
        )
    except argparse.ArgumentError:
        pass
    # `flush` may already be added
    try:
        parser.add_argument(
            "--flush",
            required=False,
            default=False,
            action="store_true",
            help="Flush every print statement.",
        )
    except argparse.ArgumentError:
        pass
    return


@function_timer
def simulate_atmosphere(args, comm, data, mc, cache_name=None, verbose=True):
    """Simulate atmospheric signal and add it to `cache_name`."""
    if not args.simulate_atmosphere:
        return
    log = Logger.get()
    timer = Timer()
    timer.start()
    if comm.world_rank == 0:
        log.info("Simulating atmosphere")
        if args.atm_cache and not os.path.isdir(args.atm_cache):
            try:
                os.makedirs(args.atm_cache)
            except FileExistsError:
                pass

    write_debug = False
    if args.atm_verbosity > 1:
        write_debug = True

    # Simulate the atmosphere signal

    atm = OpSimAtmosphere(
        out=cache_name,
        realization=2 * mc,
        lmin_center=args.atm_lmin_center,
        lmin_sigma=args.atm_lmin_sigma,
        lmax_center=args.atm_lmax_center,
        lmax_sigma=args.atm_lmax_sigma,
        gain=args.atm_gain,
        zatm=args.atm_zatm,
        zmax=args.atm_zmax,
        xstep=args.atm_xstep,
        ystep=args.atm_ystep,
        zstep=args.atm_zstep,
        nelem_sim_max=args.atm_nelem_sim_max,
        z0_center=args.atm_z0_center,
        z0_sigma=args.atm_z0_sigma,
        apply_flags=args.atm_apply_flags,
        common_flag_mask=args.common_flag_mask,
        cachedir=args.atm_cache,
        wind_dist=args.atm_wind_dist,
        write_debug=write_debug,
    )
    atm.exec(data)

    if args.simulate_coarse_atmosphere:
        atm = OpSimAtmosphere(
            out=cache_name,
            realization=2 * mc + 1,
            lmin_center=300,
            lmin_sigma=0,
            lmax_center=10000,
            lmax_sigma=0,
            gain=args.atm_gain_coarse,
            zatm=args.atm_zatm,
            zmax=args.atm_zmax,
            xstep=50,
            ystep=50,
            zstep=50,
            nelem_sim_max=30000,
            z0_center=args.atm_z0_center,
            z0_sigma=args.atm_z0_sigma,
            apply_flags=args.atm_apply_flags,
            common_flag_mask=args.common_flag_mask,
            cachedir=args.atm_cache,
            wind_dist=10000,
            write_debug=write_debug,
        )
        atm.exec(data)

    if comm.comm_world is not None:
        comm.comm_world.barrier()
    if comm.world_rank == 0:
        timer.report_clear("Atmosphere simulation")
    return


@function_timer
def scale_atmosphere_by_frequency(
    args, comm, data, freq=None, mc=0, cache_name=None, verbose=True
):
    """Scale atmospheric fluctuations by frequency.

    Assume that cached signal under totalname_freq is pure atmosphere
    and scale the absorption coefficient according to the frequency.

    If the focalplane is included in the observation and defines
    bandpasses for the detectors, the scaling is computed for each
    detector separately and `freq` is ignored.

    """
    if not args.simulate_atmosphere:
        return

    log = Logger.get()
    if comm.world_rank == 0 and verbose:
        log.info("Scaling atmosphere by frequency")
    timer = Timer()
    timer.start()
    for obs in data.obs:
        tod = obs["tod"]
        todcomm = tod.mpicomm
        site_id = obs["site_id"]
        weather = obs["weather"]
        if "focalplane" in obs:
            focalplane = obs["focalplane"]
        else:
            focalplane = None
        start_time = obs["start_time"]
        weather.set(site_id, mc, start_time)
        altitude = obs["altitude"]
        air_temperature = weather.air_temperature
        surface_pressure = weather.surface_pressure
        pwv = weather.pwv
        # Use the entire processing group to sample the absorption
        # coefficient as a function of frequency
        freqmin = 0
        if freq is None:
            freqmax = 1000
        else:
            freqmax = 2 * freq
        nfreq = 1001
        freqstep = (freqmax - freqmin) / (nfreq - 1)
        if todcomm is None:
            nfreq_task = nfreq
            my_ifreq_min = 0
            my_ifreq_max = nfreq
        else:
            nfreq_task = int(nfreq // todcomm.size) + 1
            my_ifreq_min = nfreq_task * todcomm.rank
            my_ifreq_max = min(nfreq, nfreq_task * (todcomm.rank + 1))
        my_nfreq = my_ifreq_max - my_ifreq_min
        if my_nfreq > 0:
            if atm_available_utils:
                my_freqs = freqmin + np.arange(my_ifreq_min, my_ifreq_max) * freqstep
                my_absorption = atm_absorption_coefficient_vec(
                    altitude,
                    air_temperature,
                    surface_pressure,
                    pwv,
                    my_freqs[0],
                    my_freqs[-1],
                    my_nfreq,
                )
                my_loading = atm_atmospheric_loading_vec(
                    altitude,
                    air_temperature,
                    surface_pressure,
                    pwv,
                    my_freqs[0],
                    my_freqs[-1],
                    my_nfreq,
                )
            else:
                raise RuntimeError(
                    "Atmosphere utilities from libaatm are not available"
                )
        else:
            my_freqs = np.array([])
            my_absorption = np.array([])
            my_loading = np.array([])
        if todcomm is None:
            freqs = my_freqs
            absorption = my_absorption
            loading = my_loading
        else:
            freqs = np.hstack(todcomm.allgather(my_freqs))
            absorption = np.hstack(todcomm.allgather(my_absorption))
            loading = np.hstack(todcomm.allgather(my_loading))
        for det in tod.local_dets:
            if "bandpass_transmission" in focalplane[det]:
                # We have full bandpasses for the detector
                bandpass_freqs = focalplane[det]["bandpass_freq_ghz"]
                bandpass = focalplane[det]["bandpass_transmission"]
            else:
                if "bandcenter_ghz" in focalplane[det]:
                    # Use detector bandpass from the focalplane
                    center = focalplane[det]["bandcenter_ghz"]
                    width = focalplane[det]["bandwidth_ghz"]
                else:
                    # Use default values for the entire focalplane
                    if freq is None:
                        raise RuntimeError(
                            "You must supply the nominal frequency if bandpasses "
                            "are not available"
                        )
                    center = freq
                    width = 0.2 * freq
                bandpass_freqs = np.array([center - width / 2, center + width / 2])
                bandpass = np.ones(2)
            # Normalize and interpolate the bandpass
            nstep = 1001
            fmin, fmax = bandpass_freqs[0], bandpass_freqs[-1]
            det_freqs = np.linspace(fmin, fmax, nstep)
            det_bandpass = np.interp(det_freqs, bandpass_freqs, bandpass)
            det_bandpass /= np.sum(det_bandpass)
            # Interpolate absorption and loading to bandpass frequencies
            absorption_det = np.interp(det_freqs, freqs, absorption)
            loading_det = np.interp(det_freqs, freqs, loading)
            # From brightness to thermodynamic units
            x = h * det_freqs * 1e9 / k / TCMB
            rj2cmb = (x / (np.exp(x / 2) - np.exp(-x / 2))) ** -2
            # Normalize to unity at 150GHz
            rj2cmb *= 0.5763279042527544
            absorption_det *= rj2cmb
            # Average across the bandpass
            absorption_det = np.sum(absorption_det * det_bandpass)
            loading_det = np.sum(loading_det * det_bandpass)
            cachename = "{}_{}".format(cache_name, det)
            ref = tod.cache.reference(cachename)
            ref *= absorption_det
            # Add loading, accounting for the observing elevation
            try:
                # Some TOD classes provide a shortcut to Az/El
                az, el = tod.read_azel(detector=det)
            except Exception as e:
                azelquat = tod.read_pntg(detector=det, azel=True)
                theta, phi = qa.to_position(azelquat)
                el = np.pi / 2 - theta
            ref += loading_det / np.sin(el)
            del ref

    if comm.comm_world is not None:
        comm.comm_world.barrier()
    timer.stop()
    if comm.world_rank == 0 and verbose:
        timer.report("Atmosphere scaling")
    return


@function_timer
def update_atmospheric_noise_weights(args, comm, data, freq, mc, verbose=False):
    """Update atmospheric noise weights.

    Estimate the atmospheric noise level from weather parameters and
    encode it as a noise_scale in the observation.  Madam will apply
    the noise_scale to the detector weights.  This approach assumes
    that the atmospheric noise dominates over detector noise.  To be
    more precise, we would have to add the squared noise weights but
    we do not have their relative calibration.

    """
    if not args.simulate_atmosphere:
        for obs in data.obs:
            obs["noise_scale"] = 1.0
        return
    log = Logger.get()
    if comm.world_rank == 0 and verbose:
        log.info("Updating atmospheric noise weights")
    timer = Timer()
    timer.start()
    if atm_available_utils:
        for obs in data.obs:
            site_id = obs["site_id"]
            weather = obs["weather"]
            start_time = obs["start_time"]
            weather.set(site_id, mc, start_time)
            altitude = obs["altitude"]
            absorption = atm_absorption_coefficient(
                altitude,
                weather.air_temperature,
                weather.surface_pressure,
                weather.pwv,
                freq,
            )
            obs["noise_scale"] = absorption * weather.air_temperature
    else:
        raise RuntimeError("Atmosphere utilities from libaatm are not available")

    if comm.comm_world is not None:
        comm.comm_world.barrier()
    timer.stop()
    if comm.world_rank == 0 and verbose:
        timer.report("Atmosphere weighting")
    return



# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

import os

import numpy as np

from .mpi import MPITestCase

from ..tod import Noise, sim_noise_timestream, AnalyticNoise, OpSimNoise
from ..todmap import TODHpixSpiral

from .. import rng as rng

from ._helpers import (
    create_outdir,
    create_distdata,
    boresight_focalplane,
    uniform_chunks,
)


class OpSimNoiseTest(MPITestCase):
    def setUp(self):
        fixture_name = os.path.splitext(os.path.basename(__file__))[0]
        self.outdir = create_outdir(self.comm, fixture_name)

        # Create one observation per group, and each observation will have
        # a fixed number of detectors and one chunk per process.

        # We create two data sets- one for testing uncorrelated noise and
        # one for testing correlated noise.

        self.data = create_distdata(self.comm, obs_per_group=1)
        self.data_corr = create_distdata(self.comm, obs_per_group=1)

        self.ndet = 4
        self.rate = 20.0

        # Create detectors with a range of knee frequencies.
        (
            dnames,
            dquat,
            depsilon,
            drate,
            dnet,
            dfmin,
            dfknee,
            dalpha,
        ) = boresight_focalplane(
            self.ndet,
            samplerate=self.rate,
            net=10.0,
            fmin=1.0e-5,
            fknee=np.linspace(0.0, 0.1, num=self.ndet),
        )

        # Total samples per observation
        self.totsamp = 200000

        # Chunks
        chunks = uniform_chunks(self.totsamp, nchunk=self.data.comm.group_size)

        # Noise sim oversampling
        self.oversample = 2

        # MCs for testing statistics of simulated noise
        self.nmc = 100

        # Populate the observations (one per group)

        tod = TODHpixSpiral(
            self.data.comm.comm_group,
            dquat,
            self.totsamp,
            detranks=1,
            firsttime=0.0,
            rate=self.rate,
            nside=512,
            sampsizes=chunks,
        )

        # Construct an uncorrelated analytic noise model for the detectors

        nse = AnalyticNoise(
            rate=drate,
            fmin=dfmin,
            detectors=dnames,
            fknee=dfknee,
            alpha=dalpha,
            NET=dnet,
        )

        self.data.obs[0]["tod"] = tod
        self.data.obs[0]["noise"] = nse

        # Construct a correlated analytic noise model for the detectors

        corr_freqs = {
            "noise_{}".format(x): nse.freq(dnames[x]) for x in range(self.ndet)
        }

        corr_psds = {"noise_{}".format(x): nse.psd(dnames[x]) for x in range(self.ndet)}

        corr_indices = {"noise_{}".format(x): 100 + x for x in range(self.ndet)}

        corr_mix = dict()
        for x in range(self.ndet):
            dmix = np.random.uniform(low=-1.0, high=1.0, size=self.ndet)
            corr_mix[dnames[x]] = {
                "noise_{}".format(y): dmix[y] for y in range(self.ndet)
            }

        nse_corr = Noise(
            detectors=dnames,
            freqs=corr_freqs,
            psds=corr_psds,
            mixmatrix=corr_mix,
            indices=corr_indices,
        )

        self.data_corr.obs[0]["tod"] = tod
        self.data_corr.obs[0]["noise"] = nse_corr

        return

    def test_gauss(self):
        # Test that the same samples from different calls are reproducible.
        # All processes run this identical test.

        detindx = self.ndet - 1
        telescope = 5
        realization = 1000
        component = 3
        obsindx = 1234

        key1 = realization * 4294967296 + telescope * 65536 + component
        key2 = obsindx * 4294967296 + detindx
        counter1 = 0

        compoff = 500
        ncomp = 10

        counter2 = 0
        data1 = rng.random(
            1000, sampler="gaussian", key=(key1, key2), counter=(counter1, counter2)
        )
        # print(data1[compoff:compoff+ncomp])

        counter2 = 100
        data2 = rng.random(
            1000, sampler="gaussian", key=(key1, key2), counter=(counter1, counter2)
        )
        # print(data2[compoff-100:compoff-100+ncomp])

        counter2 = 300
        data3 = rng.random(
            1000, sampler="gaussian", key=(key1, key2), counter=(counter1, counter2)
        )
        # print(data3[compoff-300:compoff-300+ncomp])

        np.testing.assert_array_almost_equal(
            data1[compoff : compoff + ncomp],
            data2[compoff - 100 : compoff - 100 + ncomp],
        )
        np.testing.assert_array_almost_equal(
            data1[compoff : compoff + ncomp],
            data3[compoff - 300 : compoff - 300 + ncomp],
        )
        return

    def test_sim(self):
        # Test the uncorrelated noise generation.

        # Verify that the white noise part of the spectrum is normalized
        # correctly.

        # We have purposely distributed the TOD data so that every process has
        # a single stationary interval for all detectors.

        rank = 0
        if self.comm is not None:
            rank = self.comm.rank

        for ob in self.data.obs:
            tod = ob["tod"]
            nse = ob["noise"]

            for det in tod.local_dets:
                fsamp = nse.rate(det)
                cutoff = 0.95 * (fsamp / 2.0)
                indx = np.where(nse.freq(det) > cutoff)

                net = nse.NET(det)
                avg = np.mean(nse.psd(det)[indx])
                netsq = net * net
                # print("det {} NETsq = {}, average white noise level = {}"
                #      "".format(det, netsq, avg))
                self.assertTrue((np.absolute(avg - netsq) / netsq) < 0.02)

            if rank == 0:
                # One process dumps debugging info
                import matplotlib.pyplot as plt

                for det in tod.local_dets:
                    savefile = os.path.join(
                        self.outdir, "out_test_simnoise_rawpsd_{}.txt".format(det)
                    )
                    np.savetxt(
                        savefile,
                        np.transpose([nse.freq(det), nse.psd(det)]),
                        delimiter=" ",
                    )

                    fig = plt.figure(figsize=(12, 8), dpi=72)
                    ax = fig.add_subplot(1, 1, 1, aspect="auto")
                    ax.loglog(
                        nse.freq(det),
                        nse.psd(det),
                        marker="o",
                        c="red",
                        label="{}: rate={:0.1f} NET={:0.1f} fknee={:0.4f}, "
                        "fmin={:0.4f}".format(
                            det,
                            nse.rate(det),
                            nse.NET(det),
                            nse.fknee(det),
                            nse.fmin(det),
                        ),
                    )

                    cur_ylim = ax.get_ylim()
                    ax.set_ylim([0.001 * (nse.NET(det) ** 2), 10.0 * cur_ylim[1]])
                    ax.legend(loc=1)
                    plt.title("Simulated PSD from toast.AnalyticNoise")

                    savefile = os.path.join(
                        self.outdir, "out_test_simnoise_rawpsd_{}.png".format(det)
                    )
                    plt.savefig(savefile)
                    plt.close()

            ntod = tod.local_samples[1]

            # this replicates the calculation in sim_noise_timestream()

            fftlen = 2
            while fftlen <= (self.oversample * ntod):
                fftlen *= 2

            freqs = {}
            psds = {}
            psdnorm = {}
            todvar = {}

            cfftlen = 2
            while cfftlen <= ntod:
                cfftlen *= 2

            # print("fftlen = ", fftlen)
            # print("cfftlen = ", cfftlen)

            checkpsd = {}
            binsamps = cfftlen // 4096
            nbins = binsamps - 1
            bstart = (self.rate / 2) / nbins
            bins = np.linspace(bstart, self.rate / 2, num=(nbins - 1), endpoint=True)
            # print("nbins = ",nbins)
            # print(bins)

            checkfreq = np.fft.rfftfreq(cfftlen, d=1 / self.rate)
            # print("checkfreq len = ",len(checkfreq))
            # print(checkfreq[:10])
            # print(checkfreq[-10:])
            checkbinmap = np.searchsorted(bins, checkfreq, side="left")
            # print("checkbinmap len = ",len(checkbinmap))
            # print(checkbinmap[:10])
            # print(checkbinmap[-10:])
            bcount = np.bincount(checkbinmap)
            # print("bcount len = ",len(bcount))
            # print(bcount)

            bintruth = {}

            idet = 0
            for det in tod.local_dets:

                dfreq = nse.rate(det) / float(fftlen)

                (pytod, freqs[det], psds[det]) = sim_noise_timestream(
                    0,
                    0,
                    0,
                    0,
                    idet,
                    nse.rate(det),
                    0,
                    ntod,
                    self.oversample,
                    nse.freq(det),
                    nse.psd(det),
                    py=True,
                )

                libtod = sim_noise_timestream(
                    0,
                    0,
                    0,
                    0,
                    idet,
                    nse.rate(det),
                    0,
                    ntod,
                    self.oversample,
                    nse.freq(det),
                    nse.psd(det),
                    py=False,
                )

                np.testing.assert_array_almost_equal(pytod, libtod, decimal=2)

                # Factor of 2 comes from the negative frequency values.
                psdnorm[det] = 2.0 * np.sum(psds[det] * dfreq)
                # print("psd[{}] integral = {}".format(det, psdnorm[det]))

                todvar[det] = np.zeros(self.nmc, dtype=np.float64)
                checkpsd[det] = np.zeros((nbins - 1, self.nmc), dtype=np.float64)

                idet += 1

            if rank == 0:
                import matplotlib.pyplot as plt

                for det in tod.local_dets:
                    savefile = os.path.join(
                        self.outdir, "out_test_simnoise_psd_{}.txt".format(det)
                    )
                    np.savetxt(
                        savefile, np.transpose([freqs[det], psds[det]]), delimiter=" "
                    )

                    fig = plt.figure(figsize=(12, 8), dpi=72)
                    ax = fig.add_subplot(1, 1, 1, aspect="auto")
                    ax.loglog(
                        freqs[det],
                        psds[det],
                        marker="+",
                        c="blue",
                        label="{}: rate={:0.1f} NET={:0.1f} fknee={:0.4f}, "
                        "fmin={:0.4f}".format(
                            det,
                            nse.rate(det),
                            nse.NET(det),
                            nse.fknee(det),
                            nse.fmin(det),
                        ),
                    )

                    cur_ylim = ax.get_ylim()
                    ax.set_ylim([0.001 * (nse.NET(det) ** 2), 10.0 * cur_ylim[1]])
                    ax.legend(loc=1)
                    plt.title(
                        "Interpolated PSD with High-pass from {:0.1f} "
                        "second Simulation Interval".format((float(ntod) / self.rate))
                    )

                    savefile = os.path.join(
                        self.outdir, "out_test_simnoise_psd_{}.png".format(det)
                    )
                    plt.savefig(savefile)
                    plt.close()

                    tmap = np.searchsorted(bins, freqs[det], side="left")
                    tcount = np.bincount(tmap)
                    tpsd = np.bincount(tmap, weights=psds[det])
                    good = tcount > 0
                    tpsd[good] /= tcount[good]
                    bintruth[det] = tpsd

            hpy = None
            if rank == 0:
                if "TOAST_TEST_BIGTOD" in os.environ.keys():
                    try:
                        import h5py as hpy
                    except ImportError:
                        # just write the first realization as usual
                        hpy = None

            # if we have the h5py module and a special environment variable is set,
            # then process zero will dump out its full timestream data for more
            # extensive sharing / tests.  Just dump a few detectors to keep the
            # file size reasonable.

            hfile = None
            dset = {}
            if hpy is not None:
                hfile = hpy.File(
                    os.path.join(self.outdir, "out_test_simnoise_tod.hdf5"), "w"
                )
                for det in tod.detectors:
                    dset[det] = hfile.create_dataset(
                        det, (self.nmc, ntod), dtype="float64"
                    )

            # Run both the numpy FFT case and the toast FFT case.

            for realization in range(self.nmc):

                # generate timestreams

                opnoise = OpSimNoise(realization=realization)
                opnoise.exec(self.data)

                if realization == 0:
                    # write timestreams to disk for debugging

                    if rank == 0:
                        import matplotlib.pyplot as plt

                        for det in tod.local_dets:

                            check = tod.cache.reference("noise_{}".format(det))

                            savefile = os.path.join(
                                self.outdir,
                                "out_test_simnoise_tod_mc0_{}.txt" "".format(det),
                            )
                            np.savetxt(savefile, np.transpose([check]), delimiter=" ")

                            fig = plt.figure(figsize=(12, 8), dpi=72)
                            ax = fig.add_subplot(1, 1, 1, aspect="auto")
                            ax.plot(
                                np.arange(len(check)),
                                check,
                                c="black",
                                label="Det {}".format(det),
                            )
                            ax.legend(loc=1)
                            plt.title(
                                "First Realization of Simulated TOD "
                                "from toast.sim_noise_timestream()"
                            )

                            savefile = os.path.join(
                                self.outdir,
                                "out_test_simnoise_tod_mc0_{}.png" "".format(det),
                            )
                            plt.savefig(savefile)
                            plt.close()

                for det in tod.local_dets:
                    # compute the TOD variance
                    ref = tod.cache.reference("noise_{}".format(det))
                    dclevel = np.mean(ref)
                    variance = np.vdot(ref - dclevel, ref - dclevel) / ntod
                    todvar[det][realization] = variance

                    if hfile is not None:
                        if det in dset:
                            dset[det][realization, :] = ref[:]

                    # compute the PSD
                    buffer = np.zeros(cfftlen, dtype=np.float64)
                    offset = (cfftlen - len(ref)) // 2
                    buffer[offset : offset + len(ref)] = ref
                    rawpsd = np.fft.rfft(buffer)
                    norm = 1.0 / (self.rate * ntod)
                    rawpsd = norm * np.abs(rawpsd ** 2)
                    bpsd = np.bincount(checkbinmap, weights=rawpsd)
                    good = bcount > 0
                    bpsd[good] /= bcount[good]
                    checkpsd[det][:, realization] = bpsd[:]

                tod.cache.clear()

            if hfile is not None:
                hfile.close()

            if rank == 0:
                np.savetxt(
                    os.path.join(self.outdir, "out_test_simnoise_tod_var.txt"),
                    np.transpose([todvar[x] for x in tod.local_dets]),
                    delimiter=" ",
                )

            if rank == 0:
                import matplotlib.pyplot as plt

                for det in tod.local_dets:
                    savefile = os.path.join(
                        self.outdir, "out_test_simnoise_tod_var_{}.txt".format(det)
                    )
                    np.savetxt(savefile, np.transpose([todvar[det]]), delimiter=" ")

                    sig = np.mean(todvar[det]) * np.sqrt(2.0 / (ntod - 1))
                    histrange = 5.0 * sig
                    histmin = psdnorm[det] - histrange
                    histmax = psdnorm[det] + histrange

                    fig = plt.figure(figsize=(12, 8), dpi=72)

                    ax = fig.add_subplot(1, 1, 1, aspect="auto")
                    plt.hist(
                        todvar[det],
                        10,
                        range=(histmin, histmax),
                        facecolor="magenta",
                        alpha=0.75,
                        label="{}:  PSD integral = {:0.1f} expected sigma = "
                        "{:0.1f}".format(det, psdnorm[det], sig),
                    )
                    ax.legend(loc=1)
                    plt.title(
                        "Distribution of TOD Variance for {} "
                        "Realizations".format(self.nmc)
                    )

                    savefile = os.path.join(
                        self.outdir, "out_test_simnoise_tod_var_{}.png".format(det)
                    )
                    plt.savefig(savefile)
                    plt.close()

                    meanpsd = np.asarray(
                        [np.mean(checkpsd[det][x, :]) for x in range(nbins - 1)]
                    )

                    fig = plt.figure(figsize=(12, 8), dpi=72)

                    ax = fig.add_subplot(1, 1, 1, aspect="auto")
                    ax.plot(bins, bintruth[det], c="k", label="Input Truth")
                    ax.plot(bins, meanpsd, c="b", marker="o", label="Mean Binned PSD")
                    ax.scatter(
                        np.repeat(bins, self.nmc),
                        checkpsd[det].flatten(),
                        marker="x",
                        color="r",
                        label="Binned PSD",
                    )
                    # ax.set_xscale("log")
                    # ax.set_yscale("log")
                    ax.legend(loc=1)
                    plt.title(
                        "Detector {} Binned PSDs for {} Realizations"
                        "".format(det, self.nmc)
                    )

                    savefile = os.path.join(
                        self.outdir, "out_test_simnoise_binpsd_dist_{}.png".format(det)
                    )
                    plt.savefig(savefile)
                    plt.close()

                    # The data will likely not be gaussian distributed.
                    # Just check that the mean is "close enough" to the truth.
                    errest = np.absolute(np.mean((meanpsd - tpsd) / tpsd))
                    # print("Det {} avg rel error = {}".format(det, errest), flush=True)
                    if nse.fknee(det) < 0.1:
                        self.assertTrue(errest < 0.1)

            # Verify that Parseval's theorem holds- that the variance of
            # the TOD equals the integral of the PSD.  We do this for an
            # ensemble of realizations
            #
            # and compare the TOD variance to the integral of the PSD
            # accounting for the error on the variance due to finite
            # numbers of samples.
            #

            for det in tod.local_dets:
                sig = np.mean(todvar[det]) * np.sqrt(2.0 / (ntod - 1))
                over3sig = np.where(
                    np.absolute(todvar[det] - psdnorm[det]) > 3.0 * sig
                )[0]
                overfrac = float(len(over3sig)) / self.nmc
                # print(det, " : ", overfrac, flush=True)
                if nse.fknee(det) < 0.01:
                    self.assertTrue(overfrac < 0.1)
        return

    def test_sim_correlated(self):
        # Test the correlated noise generation.
        opnoise = OpSimNoise(realization=0)
        opnoise.exec(self.data_corr)

        total = None

        for ob in self.data.obs:
            tod = ob["tod"]
            for det in tod.local_dets:
                # compute the TOD variance
                ref = tod.cache.reference("noise_{}".format(det))
                self.assertTrue(np.std(ref) > 0)
                if total is None:
                    total = ref.copy()
                else:
                    total[:] += ref
                del ref

        # np.testing.assert_almost_equal(np.std(total), 0)
        return



# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

from .mpi import MPITestCase

import os

import numpy as np

from ..tod import AnalyticNoise, OpSimNoise
from ..todmap import OpGroundFilter, TODGround

from ._helpers import create_outdir, create_distdata, boresight_focalplane


class OpGroundFilterTest(MPITestCase):
    def setUp(self):
        fixture_name = os.path.splitext(os.path.basename(__file__))[0]
        self.outdir = create_outdir(self.comm, fixture_name)

        # One observation per group
        self.data = create_distdata(self.comm, obs_per_group=1)

        # Detector properties.  We place one detector per process at the
        # boresight with evenly spaced polarization orientations.
        self.ndet = self.data.comm.group_size
        self.NET = 10.0
        self.rate = 20.0

        # Create detectors with a range of knee frequencies.
        (
            dnames,
            dquat,
            depsilon,
            drate,
            dnet,
            dfmin,
            dfknee,
            dalpha,
        ) = boresight_focalplane(
            self.ndet,
            samplerate=self.rate,
            net=self.NET,
            fmin=1.0e-5,
            fknee=np.linspace(0.0, 0.1, num=self.ndet),
        )

        # Samples per observation
        self.totsamp = 100000

        # bin size
        self.order = 5

        # Populate the observation
        tod = TODGround(
            self.data.comm.comm_group,
            dquat,
            self.totsamp,
            detranks=self.data.comm.group_size,
            firsttime=0.0,
            rate=self.rate,
            azmin=45,
            azmax=55,
            el=45,
        )

        # construct an analytic noise model
        nse = AnalyticNoise(
            rate=drate,
            fmin=dfmin,
            detectors=dnames,
            fknee=dfknee,
            alpha=dalpha,
            NET=dnet,
        )

        self.data.obs[0]["tod"] = tod
        self.data.obs[0]["noise"] = nse

    def test_filter(self):
        # generate timestreams
        op = OpSimNoise()
        op.exec(self.data)

        # Replace the noise with a ground-synchronous signal
        old_rms = []
        for ob in self.data.obs:
            tod = ob["tod"]
            az = tod.read_boresight_az()
            orms = {}
            for det in tod.local_dets:
                cachename = "noise_{}".format(det)
                y = tod.cache.reference(cachename)
                y[:] = np.sin(az)
                orms[det] = np.std(y)
                del y
            old_rms.append(orms)

        # Filter timestreams
        op = OpGroundFilter(name="noise", filter_order=self.order, common_flag_mask=0)
        op.exec(self.data)

        # Ensure all timestreams are zeroed out by the filter
        for ob, orms in zip(self.data.obs, old_rms):
            tod = ob["tod"]
            for det in tod.local_dets:
                cachename = "noise_{}".format(det)
                y = tod.cache.reference(cachename)
                rms = np.std(y)
                old = orms[det]
                if np.abs(rms / old) > 1.0e-3:
                    raise RuntimeError(
                        "det {} old rms = {}, new rms = {}".format(det, old, rms)
                    )
                del y
        return

    def test_filter_split(self):
        # generate timestreams
        op = OpSimNoise()
        op.exec(self.data)

        # Replace the noise with a ground-synchronous signal
        old_rms = []
        for ob in self.data.obs:
            tod = ob["tod"]
            az = tod.read_boresight_az()
            orms = {}
            for det in tod.local_dets:
                cachename = "noise_{}".format(det)
                y = tod.cache.reference(cachename)
                y[:] = np.sin(az)
                orms[det] = np.std(y)
                del y
            old_rms.append(orms)

        # Filter timestreams
        op = OpGroundFilter(
            name="noise",
            filter_order=self.order,
            common_flag_mask=0,
            split_template=True,
            trend_order=1,
            detrend=True,
        )
        op.exec(self.data)

        # Ensure all timestreams are zeroed out by the filter
        for ob, orms in zip(self.data.obs, old_rms):
            tod = ob["tod"]
            for det in tod.local_dets:
                cachename = "noise_{}".format(det)
                y = tod.cache.reference(cachename)
                rms = np.std(y)
                old = orms[det]
                if np.abs(rms / old) > 1.0e-3:
                    raise RuntimeError(
                        "det {} old rms = {}, new rms = {}".format(det, old, rms)
                    )
                del y
        return



# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

import numpy as np

import healpy as hp

from ..mpi import MPI

from ..timing import function_timer

from ..utils import Logger, Timer

from ..op import Operator

from .pysm import pysm

if pysm is not None:
    import pysm.units as u
    from .pysm import PySMSky

from ..map import DistPixels

from .sim_det_map import OpSimScan


def extract_local_dets(data):
    """Extracts the local detectors from the TOD objects

    Some detectors could only appear in some observations, so we need
    to loop through all observations and accumulate all detectors in
    a set
    """
    local_dets = set()
    for obs in data.obs:
        tod = obs["tod"]
        local_dets.update(tod.local_dets)
    return local_dets


@function_timer
def assemble_map_on_rank0(comm, local_map, pixel_indices, n_components, npix):
    if comm is None:
        return local_map
    else:
        full_maps_rank0 = (
            np.zeros((n_components, npix), dtype=np.float64) if comm.rank == 0 else None
        )
        local_map_buffer = np.zeros((n_components, npix), dtype=np.float64)
        local_map_buffer[:, pixel_indices] = local_map
        comm.Reduce(local_map_buffer, full_maps_rank0, root=0, op=MPI.SUM)
        return full_maps_rank0


@function_timer
def extract_detector_parameters(det, focalplanes):
    for fp in focalplanes:
        if det in fp:
            if "fwhm" in fp[det]:
                return (
                    fp[det]["bandcenter_ghz"],
                    fp[det]["bandwidth_ghz"],
                    fp[det]["fwhm"] / 60,
                )
            else:
                return fp[det]["bandcenter_ghz"], fp[det]["bandwidth_ghz"], -1
    raise RuntimeError("Cannot find detector {} in any focalplane")


class OpSimPySM(Operator):
    """Operator which generates a bandpass integrated and smoothed sky signal with PySM 3

    This operator:
    * Extracts band centers,  bandwidths and fwhm from the defined focalplane
    * Creates a `PySMSky` object
    * Runs `PySMSky` and gets distributed maps
    * Performs distributed smoothing with libsharp facilities provided by PySM 3
    * Communicates the distributed map to the first process
    * Communicates to each of the processes their local pixels
    * Rescans the pixels to a timeline

    For PySM related arguments, see the PySMSky docstring

    Args:
        data (toast.Data) : Data object with pixelization metadata
        comm (mpi4py.MPI.Comm): MPI communicator
        out (str): accumulate data to the cache with name <out>_<detector>.
            If the named cache objects do not exist, then they are created.
        focalplanes (list(dict)): List of focalplanes dictionaries with channel
            name as key, another dictionary with keys "bandcenter_ghz", "bandwidth_ghz",
            "fmin", "fwhm" (in arcmin)
        apply_beam (bool): Whether to perform gaussian smoothing with libsharp using the
            fwhm defined in the focalplane
        nest (bool): HEALPix nest or ring pixels
        units (str): Output units.
        debug (bool):  Verbose progress reports.
        coord (str): Output reference frame
        pixels (str) :  Cache prefix for precomputed pixels
    """

    @function_timer
    def __init__(
        self,
        data,
        comm=None,
        out="signal",
        pysm_model=None,
        pysm_precomputed_cmb_K_CMB=None,
        pysm_component_objects=None,
        focalplanes=None,
        apply_beam=False,
        nest=True,
        units="K_CMB",
        debug=False,
        coord="G",
        map_dist=None,
        pixels="pixels",
    ):
        if pysm is None:
            raise RuntimeError("PySM not available")
        # Call the parent class constructor.
        super().__init__()
        self._out = out
        self._nest = nest
        self.comm = comm
        self._debug = debug
        self.pysm_precomputed_cmb_K_CMB = pysm_precomputed_cmb_K_CMB
        self.coord = coord
        self.pixels = pixels
        self.npix = data["{}_npix".format(self.pixels)]
        self.nside = hp.npix2nside(self.npix)

        self.pysm_sky = PySMSky(
            comm=self.comm,
            pixel_indices=None,
            nside=self.nside,
            pysm_sky_config=pysm_model,
            pysm_component_objects=pysm_component_objects,
            pysm_precomputed_cmb_K_CMB=self.pysm_precomputed_cmb_K_CMB,
            units=units,
            map_dist=map_dist,
        )

        self.focalplanes = focalplanes
        self.distmap = DistPixels(
            data,
            comm=comm,
            nnz=3,
            dtype=np.float32,
            pixels=self.pixels,
        )
        self.apply_beam = apply_beam

    @function_timer
    def exec(self, data):
        log = Logger.get()
        rank = 0
        if self.comm is not None:
            rank = self.comm.rank

        local_dets = extract_local_dets(data)
        bandpasses = {}
        fwhm_deg = {}
        N_POINTS_BANDPASS = 10  # possibly take as parameter
        for det in local_dets:
            bandcenter, bandwidth, fwhm_deg[det] = extract_detector_parameters(
                det, self.focalplanes
            )
            bandpasses[det] = (
                np.linspace(
                    bandcenter - bandwidth / 2,
                    bandcenter + bandwidth / 2,
                    N_POINTS_BANDPASS,
                ),
                np.ones(N_POINTS_BANDPASS),
            )

        if rank == 0:
            log.debug("Collecting, Broadcasting map")
        tm = Timer()
        tm.start()

        for det in local_dets:
            local_maps = dict()
            if self.comm is not None:
                self.comm.Barrier()
            if rank == 0:
                log.debug("Running PySM on {}".format(det))
            self.pysm_sky.exec(local_maps, out="sky", bandpasses={det: bandpasses[det]})

            if self.apply_beam:
                if fwhm_deg[det] == -1:
                    raise RuntimeError(
                        "OpSimPySM: apply beam is True but focalplane doesn't "
                        "have fwhm"
                    )
                # LibSharp also supports transforming multiple channels
                # together each with own beam
                if self.comm is not None:
                    self.comm.Barrier()
                if rank == 0:
                    log.debug("Executing Smoothing with libsharp on {}".format(det))
                local_maps[
                    "sky_{}".format(det)
                ] = pysm.apply_smoothing_and_coord_transform(
                    local_maps["sky_{}".format(det)],
                    fwhm=fwhm_deg[det] * u.deg,
                    map_dist=self.pysm_sky.map_dist,
                )
                if self.comm is not None:
                    self.comm.Barrier()
                if rank == 0:
                    log.debug("Smoothing completed on {}".format(det))

            n_components = 3

            if self.comm is not None:
                self.comm.Barrier()
            if rank == 0:
                log.debug(
                    "Assemble PySM map on rank0, shape of local map is {}".format(
                        local_maps["sky_{}".format(det)].shape
                    )
                )
            full_map_rank0 = assemble_map_on_rank0(
                self.comm,
                local_maps["sky_{}".format(det)],
                np.arange(len(local_maps["sky_{}".format(det)][0]))
                if self.comm is None
                else self.pysm_sky.map_dist.pixel_indices,
                n_components,
                self.npix,
            )

            if self.comm is not None:
                self.comm.Barrier()
            if rank == 0:
                log.debug("Communication completed")
            if rank == 0 and self.coord != "G":
                # PySM is always in Galactic, make rotation to Ecliptic or Equatorial
                rot = hp.Rotator(coord=["G", self.coord])
                # this requires healpy 1.12.8
                try:
                    full_map_rank0 = rot.rotate_map_alms(
                        full_map_rank0, use_pixel_weights=True
                    )
                except AttributeError:
                    print(
                        "PySM coordinate conversion from G to another reference frame requires"
                        "healpy.Rotator.rotate_map_alms available since healpy 1.12.8"
                    )
                    raise
            if rank == 0 and self._nest:
                # PySM is RING, convert to NEST if desired.
                full_map_rank0 = hp.reorder(full_map_rank0, r2n=True)
            # full_map_rank0 dict contains on rank 0 the smoothed PySM map

            if self.comm is not None:
                self.comm.Barrier()
            if rank == 0:
                log.debug(
                    "PySM map min / max pixel value = {} / {}".format(
                        hp.ma(full_map_rank0).min(), hp.ma(full_map_rank0).max()
                    )
                )
                log.debug("Broadcasting the map to other processes")
            self.distmap.broadcast_healpix_map(full_map_rank0)
            if rank == 0:
                log.debug("Running OpSimScan")
            scansim = OpSimScan(input_map=self.distmap, out=self._out, dets=[det])
            scansim.exec(data)
            if rank == 0:
                tod = data.obs[0]["tod"]
                sig = tod.cache.reference(self._out + "_" + det)
                log.debug(
                    "Rank 0 timeline min / max after smoothing = {} / {}".format(
                        sig.min(), sig.max()
                    )
                )

        tm.stop()
        if rank == 0:
            # tm.report("PySM Operator")
            pass

        return



# This generates a small dataset with healpy
# that can be inserted into the unit tests.
# To update test data, do:
#
# %> python gen_healpix_data.py > data_healpix.cpp
#

import numpy as np
import healpy as hp


nside = 16384

angperring = 10

nring = 7

thetainc = 0.98 * (np.pi / nring)

phiinc = 0.98 * (2.0 * np.pi / angperring)

theta = np.zeros(nring * angperring)
phi = np.zeros_like(theta)

n = 0

for t in range(nring):
    for p in range(angperring):
        theta[n] = t * thetainc
        phi[n] = p * phiinc
        n += 1

pixring = hp.ang2pix(nside, theta, phi)
pixnest = hp.ring2nest(nside, pixring)

ringtheta, ringphi = hp.pix2ang(nside, pixring)

print("    int64_t nside = {};".format(nside))
print("")

print("    int64_t ntest = {};".format(n))
print("")

print("    double theta[{}] = {{".format(n))
for i in range(n):
    print("        {},".format(theta[i]))
print("    };")
print("")

print("    double phi[{}] = {{".format(n))
for i in range(n):
    print("        {},".format(phi[i]))
print("    };")
print("")

print("    int64_t pixring[{}] = {{".format(n))
for i in range(n):
    print("        {},".format(pixring[i]))
print("    };")
print("")

print("    int64_t pixnest[{}] = {{".format(n))
for i in range(n):
    print("        {},".format(pixnest[i]))
print("    };")
print("")

print("    double ringtheta[{}] = {{".format(n))
for i in range(n):
    print("        {},".format(ringtheta[i]))
print("    };")
print("")

print("    double ringphi[{}] = {{".format(n))
for i in range(n):
    print("        {},".format(ringphi[i]))
print("    };")
print("")



#!/usr/bin/env python3

# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

"""
This script runs a satellite simulation and makes a map.
"""

import os
import sys
import re
import argparse
import traceback

import pickle

import numpy as np

from astropy.io import fits

from toast.mpi import get_world, Comm

from toast.dist import distribute_uniform, Data

from toast.utils import Logger, Environment

from toast.vis import set_backend

from toast.timing import function_timer, GlobalTimers, Timer, gather_timers
from toast.timing import dump as dump_timing

from toast.tod import regular_intervals, plot_focalplane, OpApplyGain
from toast.todmap import TODSatellite, slew_precession_axis

from toast import pipeline_tools


def parse_arguments(comm, procs):
    log = Logger.get()

    parser = argparse.ArgumentParser(
        description="Simulate satellite boresight pointing and make a map.",
        fromfile_prefix_chars="@",
    )

    pipeline_tools.add_dist_args(parser)
    pipeline_tools.add_pointing_args(parser)
    pipeline_tools.add_tidas_args(parser)
    pipeline_tools.add_spt3g_args(parser)
    pipeline_tools.add_dipole_args(parser)
    pipeline_tools.add_sky_map_args(parser)
    pipeline_tools.add_pysm_args(parser)
    pipeline_tools.add_mc_args(parser)
    pipeline_tools.add_noise_args(parser)
    pipeline_tools.add_todsatellite_args(parser)
    pipeline_tools.add_conviqt_args(parser)

    parser.add_argument(
        "--outdir", required=False, default="out", help="Output directory"
    )
    parser.add_argument(
        "--debug",
        required=False,
        default=False,
        action="store_true",
        help="Write diagnostics",
    )

    pipeline_tools.add_madam_args(parser)
    pipeline_tools.add_mapmaker_args(parser)
    pipeline_tools.add_binner_args(parser)

    parser.add_argument(
        "--madam",
        required=False,
        action="store_true",
        help="Use libmadam for map-making",
        dest="use_madam",
    )
    parser.add_argument(
        "--no-madam",
        required=False,
        action="store_false",
        help="Do not use libmadam for map-making [default]",
        dest="use_madam",
    )
    parser.set_defaults(use_madam=False)

    parser.add_argument(
        "--focalplane",
        required=False,
        default=None,
        help="Pickle file containing a dictionary of detector properties.  "
        "The keys of this dict are the detector names, and each value is also "
        'a dictionary with keys "quat" (4 element ndarray), "fwhm" '
        '(float, arcmin), "fknee" (float, Hz), "alpha" (float), and "NET" '
        '(float).  For optional plotting, the key "color" can specify a '
        "valid matplotlib color string.",
    )

    parser.add_argument(
        "--gain",
        required=False,
        default=None,
        help="Calibrate the input timelines with a set of gains from a"
        "FITS file containing 3 extensions:"
        "HDU named DETECTORS : table with list of detector names in a column named DETECTORS"
        "HDU named TIME: table with common timestamps column named TIME"
        "HDU named GAINS: 2D image of floats with one row per detector and one column per value.",
    )

    try:
        args = parser.parse_args()
    except SystemExit:
        sys.exit()

    if comm.world_rank == 0:
        log.info("\n")
        log.info("All parameters:")
        for ag in vars(args):
            log.info("{} = {}".format(ag, getattr(args, ag)))
        log.info("\n")

    groupsize = args.group_size
    if groupsize is None or groupsize <= 0:
        groupsize = procs

    # This is the 2-level toast communicator.
    comm = Comm(groupsize=groupsize)

    return args, comm, groupsize


def load_focalplane(args, comm):
    """Load focalplane information"""
    timer = Timer()
    gain = None
    fp = None
    if comm.world_rank == 0:
        if args.focalplane is None:
            # in this case, create a fake detector at the boresight
            # with a pure white noise spectrum.
            fake = {}
            fake["quat"] = np.array([0.0, 0.0, 1.0, 0.0])
            fake["fwhm"] = 30.0
            fake["fknee"] = 0.0
            fake["fmin"] = 1.0e-5
            fake["alpha"] = 1.0
            fake["NET"] = 1.0
            fake["polangle_deg"] = 0
            fake["color"] = "r"
            fp = {}
            fp["bore"] = fake
        else:
            with open(args.focalplane, "rb") as p:
                fp = pickle.load(p)

        if args.gain is not None:
            gain = {}
            with fits.open(args.gain) as f:
                gain["TIME"] = np.array(f["TIME"].data["TIME"])
                for i_det, det_name in f["DETECTORS"].data["DETECTORS"]:
                    gain[det_name] = np.array(f["GAINS"].data[i_det, :])

    if comm.comm_world is not None:
        if args.gain is not None:
            gain = comm.comm_world.bcast(gain, root=0)
        fp = comm.comm_world.bcast(fp, root=0)

    timer.stop()
    if comm.world_rank == 0:
        timer.report("Create focalplane ({} dets)".format(len(fp.keys())))

    if args.debug:
        if comm.world_rank == 0:
            outfile = os.path.join(args.outdir, "focalplane.png")
            set_backend()
            dquats = {x: fp[x]["quat"] for x in fp.keys()}
            dfwhm = {x: fp[x]["fwhm"] for x in fp.keys()}
            plot_focalplane(dquats, 10.0, 10.0, outfile, fwhm=dfwhm)

    # For purposes of this simulation, we use detector noise
    # weights based on the NET (white noise level).  If the destriping
    # baseline is too long, this will not be the best choice.

    detweights = {}
    for d in fp.keys():
        net = fp[d]["NET"]
        detweights[d] = 1.0 / (args.sample_rate * net * net)

    return fp, gain, detweights


def create_observations(args, comm, focalplane, groupsize):
    timer = Timer()
    timer.start()
    log = Logger.get()

    if groupsize > len(focalplane.keys()):
        if comm.world_rank == 0:
            log.error("process group is too large for the number of detectors")
            comm.comm_world.Abort()

    # Detector information from the focalplane

    detectors = sorted(focalplane.keys())
    detquats = {}
    detindx = None
    if "index" in focalplane[detectors[0]]:
        detindx = {}

    for d in detectors:
        detquats[d] = focalplane[d]["quat"]
        if detindx is not None:
            detindx[d] = focalplane[d]["index"]

    # Distribute the observations uniformly

    groupdist = distribute_uniform(args.obs_num, comm.ngroups)

    # Compute global time and sample ranges of all observations

    obsrange = regular_intervals(
        args.obs_num,
        args.start_time,
        0,
        args.sample_rate,
        3600 * args.obs_time_h,
        3600 * args.gap_h,
    )

    noise = pipeline_tools.get_analytic_noise(args, comm, focalplane)

    # The distributed timestream data

    data = Data(comm)

    # Every process group creates its observations

    group_firstobs = groupdist[comm.group][0]
    group_numobs = groupdist[comm.group][1]

    for ob in range(group_firstobs, group_firstobs + group_numobs):
        tod = TODSatellite(
            comm.comm_group,
            detquats,
            obsrange[ob].samples,
            coord=args.coord,
            firstsamp=obsrange[ob].first,
            firsttime=obsrange[ob].start,
            rate=args.sample_rate,
            spinperiod=args.spin_period_min,
            spinangle=args.spin_angle_deg,
            precperiod=args.prec_period_min,
            precangle=args.prec_angle_deg,
            detindx=detindx,
            detranks=comm.group_size,
            hwprpm=args.hwp_rpm,
            hwpstep=args.hwp_step_deg,
            hwpsteptime=args.hwp_step_time_s,
        )

        obs = {}
        obs["name"] = "science_{:05d}".format(ob)
        obs["tod"] = tod
        obs["intervals"] = None
        obs["baselines"] = None
        obs["noise"] = noise
        obs["id"] = ob
        obs["focalplane"] = pipeline_tools.Focalplane(focalplane)

        data.obs.append(obs)

    if comm.world_rank == 0:
        timer.report_clear("Read parameters, compute data distribution")

    # Since we are simulating noise timestreams, we want
    # them to be contiguous and reproducible over the whole
    # observation.  We distribute data by detector within an
    # observation, so ensure that our group size is not larger
    # than the number of detectors we have.

    # we set the precession axis now, which will trigger calculation
    # of the boresight pointing.

    for ob in range(group_numobs):
        curobs = data.obs[ob]
        tod = curobs["tod"]

        # Get the global sample offset from the original distribution of
        # intervals
        obsoffset = obsrange[group_firstobs + ob].first

        # Constantly slewing precession axis
        degday = 360.0 / 365.25
        precquat = np.empty(4 * tod.local_samples[1], dtype=np.float64).reshape((-1, 4))
        slew_precession_axis(
            precquat,
            firstsamp=(obsoffset + tod.local_samples[0]),
            samplerate=args.sample_rate,
            degday=degday,
        )

        tod.set_prec_axis(qprec=precquat)
        del precquat

    if comm.world_rank == 0:
        timer.report_clear("Construct boresight pointing")

    return data


def main():
    env = Environment.get()
    log = Logger.get()
    gt = GlobalTimers.get()
    gt.start("toast_satellite_sim (total)")
    timer0 = Timer()
    timer0.start()

    mpiworld, procs, rank, comm = pipeline_tools.get_comm()
    args, comm, groupsize = parse_arguments(comm, procs)

    # Parse options

    tmr = Timer()
    tmr.start()

    if comm.world_rank == 0:
        os.makedirs(args.outdir, exist_ok=True)

    focalplane, gain, detweights = load_focalplane(args, comm)
    if comm.world_rank == 0:
        tmr.report_clear("Load focalplane")

    data = create_observations(args, comm, focalplane, groupsize)
    if comm.world_rank == 0:
        tmr.report_clear("Create observations")

    pipeline_tools.expand_pointing(args, comm, data)
    if comm.world_rank == 0:
        tmr.report_clear("Expand pointing")

    signalname = None
    if args.pysm_model:
        skyname = pipeline_tools.simulate_sky_signal(
            args, comm, data, [focalplane], "signal"
        )
    else:
        skyname = pipeline_tools.scan_sky_signal(args, comm, data, "signal")
    if skyname is not None:
        signalname = skyname
    if comm.world_rank == 0:
        tmr.report_clear("Simulate sky signal")

    # NOTE: Conviqt could use different input file names for different
    # Monte Carlo indices, but the operator would need to be invoked within
    # the Monte Carlo loop.
    skyname = pipeline_tools.apply_conviqt(
        args,
        comm,
        data,
        "signal",
        mc=args.MC_start,
    )
    if skyname is not None:
        signalname = skyname
    if comm.world_rank == 0:
        tmr.report_clear("Apply beam convolution")

    diponame = pipeline_tools.simulate_dipole(args, comm, data, "signal")
    if diponame is not None:
        signalname = diponame
    if comm.world_rank == 0:
        tmr.report_clear("Simulate dipole")

    # in debug mode, print out data distribution information
    if args.debug:
        handle = None
        if comm.world_rank == 0:
            handle = open(os.path.join(args.outdir, "distdata.txt"), "w")
        data.info(handle)
        if comm.world_rank == 0:
            handle.close()
        if comm.comm_world is not None:
            comm.comm_world.barrier()
        if comm.world_rank == 0:
            tmr.report_clear("Dumping data distribution")

    # in debug mode, print out data distribution information
    if args.debug:
        handle = None
        if comm.world_rank == 0:
            handle = open(os.path.join(args.outdir, "distdata.txt"), "w")
        data.info(handle)
        if comm.world_rank == 0:
            handle.close()
        if comm.comm_world is not None:
            comm.comm_world.barrier()
        if comm.world_rank == 0:
            tmr.report_clear("Dumping data distribution")

    # Mapmaking.

    if args.use_madam:
        # Initialize madam parameters
        madampars = pipeline_tools.setup_madam(args)
        if comm.comm_world is not None:
            comm.comm_world.barrier()
        if comm.world_rank == 0:
            tmr.report_clear("Initialize madam map-making")

    # Loop over Monte Carlos

    firstmc = args.MC_start
    nmc = args.MC_count

    for mc in range(firstmc, firstmc + nmc):
        mctmr = Timer()
        mctmr.start()

        # create output directory for this realization
        outpath = os.path.join(args.outdir, "mc_{:03d}".format(mc))

        pipeline_tools.simulate_noise(
            args, comm, data, mc, "tot_signal", overwrite=True
        )
        if comm.comm_world is not None:
            comm.comm_world.barrier()
        if comm.world_rank == 0:
            tmr.report_clear("    Simulate noise {:04d}".format(mc))

        # add sky signal
        pipeline_tools.add_signal(args, comm, data, "tot_signal", signalname)
        if comm.comm_world is not None:
            comm.comm_world.barrier()
        if comm.world_rank == 0:
            tmr.report_clear("    Add sky signal {:04d}".format(mc))

        if gain is not None:
            op_apply_gain = OpApplyGain(gain, name="tot_signal")
            op_apply_gain.exec(data)
            if comm.comm_world is not None:
                comm.comm_world.barrier()
            if comm.world_rank == 0:
                tmr.report_clear("    Apply gains {:04d}".format(mc))

        if mc == firstmc:
            # For the first realization, optionally export the
            # timestream data.  If we had observation intervals defined,
            # we could pass "use_interval=True" to the export operators,
            # which would ensure breaks in the exported data at
            # acceptable places.
            pipeline_tools.output_tidas(args, comm, data, "tot_signal")
            pipeline_tools.output_spt3g(args, comm, data, "tot_signal")
            if comm.comm_world is not None:
                comm.comm_world.barrier()
            if comm.world_rank == 0:
                tmr.report_clear("    Write TOD snapshot {:04d}".format(mc))

        if args.use_madam:
            pipeline_tools.apply_madam(
                args, comm, data, madampars, outpath, detweights, "tot_signal"
            )
        else:
            pipeline_tools.apply_mapmaker(args, comm, data, outpath, "tot_signal")

        if comm.comm_world is not None:
            comm.comm_world.barrier()
        if comm.world_rank == 0:
            tmr.report_clear("  Map-making {:04d}".format(mc))

        if comm.comm_world is not None:
            comm.comm_world.barrier()
        if comm.world_rank == 0:
            mctmr.report_clear("  Monte Carlo loop {:04d}".format(mc))

    gt.stop_all()
    if comm.comm_world is not None:
        comm.comm_world.barrier()
    tmr.stop()
    tmr.clear()
    tmr.start()
    alltimers = gather_timers(comm=comm.comm_world)
    if comm.world_rank == 0:
        out = os.path.join(args.outdir, "timing")
        dump_timing(alltimers, out)
        tmr.stop()
        tmr.report("Gather and dump timing info")
        timer0.report_clear("toast_satellite_sim.py")
    return


if __name__ == "__main__":
    try:
        main()
    except Exception:
        # We have an unhandled exception on at least one process.  Print a stack
        # trace for this process and then abort so that all processes terminate.
        mpiworld, procs, rank = get_world()
        if procs == 1:
            raise
        exc_type, exc_value, exc_traceback = sys.exc_info()
        lines = traceback.format_exception(exc_type, exc_value, exc_traceback)
        lines = ["Proc {}: {}".format(rank, x) for x in lines]
        print("".join(lines), flush=True)
        if mpiworld is not None:
            mpiworld.Abort(6)



# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

from .mpi import MPITestCase

from ..utils import AlignedU64

from ..rng import random, random_multi

import numpy as np


class RNGTest(MPITestCase):
    def setUp(self):
        # data
        self.size = 11
        self.counter = [1357111317, 888118218888]
        self.key = [3405692589, 3131965165]
        self.counter00 = [0, 0]
        self.key00 = [0, 0]
        self.nstream = 100

        # C test output with counter=[1357111317,888118218888] and
        # key=[3405692589,3131965165]
        # self.array_gaussian = np.array([-0.602799, 2.141513, -0.433604,
        # 0.493275, -0.037459, -0.926340, -0.536562, -0.064849, -0.662582,
        # -1.024292, -0.170119], np.float64)

        self.array_m11 = np.array(
            [
                -0.951008,
                0.112014,
                -0.391117,
                0.858437,
                -0.232332,
                -0.929797,
                0.513278,
                -0.722889,
                -0.439833,
                0.814677,
                0.466897,
            ],
            np.float64,
        )
        self.array_01 = np.array(
            [
                0.524496,
                0.056007,
                0.804442,
                0.429218,
                0.883834,
                0.535102,
                0.256639,
                0.638556,
                0.780084,
                0.407338,
                0.233448,
            ],
            np.float64,
        )
        self.array_uint64 = np.array(
            [
                9675248043493244317,
                1033143684219887964,
                14839328367301273822,
                7917682351778602270,
                16303863741333868668,
                9870884412429777903,
                4734154306332135586,
                11779270208507399991,
                14390002533568630569,
                7514066637753215609,
                4306362335420736255,
            ],
            np.uint64,
        )

        # C test output with counter=[0,0] and key=[0,0]
        # self.array00_gaussian = np.array([-0.680004, -0.633214, -1.523790,
        # -1.847484, -0.427139, 0.991348, 0.601200, 0.481707, -0.085967,
        # 0.110980, -1.220734], np.float64)

        self.array00_m11 = np.array(
            [
                -0.478794,
                -0.704256,
                0.533997,
                0.004571,
                0.392376,
                -0.785938,
                -0.373569,
                0.866371,
                0.325575,
                -0.266422,
                0.937621,
            ],
            np.float64,
        )
        self.array00_01 = np.array(
            [
                0.760603,
                0.647872,
                0.266998,
                0.002285,
                0.196188,
                0.607031,
                0.813215,
                0.433185,
                0.162788,
                0.866789,
                0.468810,
            ],
            np.float64,
        )
        self.array00_uint64 = np.array(
            [
                14030652003081164901,
                11951131804325250240,
                4925249918008276254,
                42156276261651215,
                3619028682724454876,
                11197741606642300638,
                15001177968947004470,
                7990859118804543502,
                3002902877118036975,
                15989435820833075781,
                8648023362736035120,
            ],
            np.uint64,
        )

    def test_rng_gaussian(self):
        # # Testing with any counter and any key
        # result = random(self.size, counter=self.counter, key=self.key)
        # self.assertTrue((result > -10).all() and (result < 10).all())
        # np.testing.assert_array_almost_equal(result, self.array_gaussian)

        # # Testing with counter=[0,0] and key=[0,0]
        # result = random(self.size, counter=self.counter00, key=self.key00)
        # self.assertTrue((result > -10).all() and (result < 10).all())
        # np.testing.assert_array_almost_equal(result, self.array00_gaussian)

        # Test reproducibility
        result1 = random(20, key=[0, 0], counter=[0, 0], sampler="gaussian")
        result2 = random(20, key=[0, 0], counter=[0, 5], sampler="gaussian")
        np.testing.assert_array_almost_equal(result1[5:], result2[:-5])

        # ...And with threads
        result1 = random(
            20, key=[0, 0], counter=[0, 0], sampler="gaussian", threads=True
        )
        result2 = random(
            20, key=[0, 0], counter=[0, 5], sampler="gaussian", threads=True
        )
        np.testing.assert_array_almost_equal(result1[5:], result2[:-5])
        return

    def test_rng_m11(self):
        # Testing with any counter and any key
        result = random(
            self.size, key=self.key, counter=self.counter, sampler="uniform_m11"
        )
        self.assertTrue((result > -1.0).all() and (result < 1.0).all())
        np.testing.assert_array_almost_equal(result, self.array_m11)

        # Testing with counter=[0,0] and key=[0,0]
        result = random(
            self.size, key=self.key00, counter=self.counter00, sampler="uniform_m11"
        )
        self.assertTrue((result > -1.0).all() and (result < 1.0).all())
        np.testing.assert_array_almost_equal(result, self.array00_m11)

        # ...And with threads
        result = random(
            self.size,
            key=self.key,
            counter=self.counter,
            sampler="uniform_m11",
            threads=True,
        )
        self.assertTrue((result > -1.0).all() and (result < 1.0).all())
        np.testing.assert_array_almost_equal(result, self.array_m11)

        result = random(
            self.size,
            key=self.key00,
            counter=self.counter00,
            sampler="uniform_m11",
            threads=True,
        )
        self.assertTrue((result > -1.0).all() and (result < 1.0).all())
        np.testing.assert_array_almost_equal(result, self.array00_m11)

        return

    def test_rng_01(self):
        # Testing with any counter and any key
        result = random(
            self.size, key=self.key, counter=self.counter, sampler="uniform_01"
        )
        self.assertTrue((result > 0.0).all() and (result < 1.0).all())
        np.testing.assert_array_almost_equal(np.array(result), self.array_01)

        # Testing with counter=[0,0] and key=[0,0]
        result = random(
            self.size, key=self.key00, counter=self.counter00, sampler="uniform_01"
        )
        self.assertTrue((result > 0.0).all() and (result < 1.0).all())
        np.testing.assert_array_almost_equal(result, self.array00_01)

        # ...And with threads
        result = random(
            self.size,
            key=self.key,
            counter=self.counter,
            sampler="uniform_01",
            threads=True,
        )
        self.assertTrue((result > 0.0).all() and (result < 1.0).all())
        np.testing.assert_array_almost_equal(result, self.array_01)

        result = random(
            self.size,
            key=self.key00,
            counter=self.counter00,
            sampler="uniform_01",
            threads=True,
        )
        self.assertTrue((result > 0.0).all() and (result < 1.0).all())
        np.testing.assert_array_almost_equal(result, self.array00_01)

        return

    def test_rng_uint64(self):
        # Testing with any counter and any key
        result = random(
            self.size, key=self.key, counter=self.counter, sampler="uniform_uint64"
        )
        self.assertTrue(type(result) == AlignedU64)
        np.testing.assert_array_equal(result, self.array_uint64)

        # Testing with counter=[0,0] and key=[0,0]
        result = random(
            self.size, key=self.key00, counter=self.counter00, sampler="uniform_uint64"
        )
        self.assertTrue(type(result) == AlignedU64)
        np.testing.assert_array_equal(result, self.array00_uint64)

        # ...And with threads
        result = random(
            self.size,
            key=self.key,
            counter=self.counter,
            sampler="uniform_uint64",
            threads=True,
        )
        self.assertTrue(type(result) == AlignedU64)
        np.testing.assert_array_equal(result, self.array_uint64)

        result = random(
            self.size,
            key=self.key00,
            counter=self.counter00,
            sampler="uniform_uint64",
            threads=True,
        )
        self.assertTrue(type(result) == AlignedU64)
        np.testing.assert_array_equal(result, self.array00_uint64)

        return

    # Multiple stream testing.

    def test_rng_gaussian_multi(self):
        # Test reproducibility
        samples = [20 for x in range(self.nstream)]
        keys = [(0, 0) for x in range(self.nstream)]
        counters = [(0, 0) for x in range(self.nstream)]

        result1 = random_multi(samples, keys, counters, sampler="gaussian")

        counters = [(0, 5) for x in range(self.nstream)]
        result2 = random_multi(samples, keys, counters, sampler="gaussian")

        for i in range(self.nstream):
            np.testing.assert_array_almost_equal(result1[i][5:], result2[i][:-5])
        return

    def test_rng_m11_multi(self):
        # Testing with any counter and any key
        samples = [self.size for x in range(self.nstream)]
        keys = [self.key for x in range(self.nstream)]
        counters = [self.counter for x in range(self.nstream)]
        result = random_multi(samples, keys, counters, sampler="uniform_m11")
        for i in range(self.nstream):
            self.assertTrue((result[i] > -1.0).all() and (result[i] < 1.0).all())
            np.testing.assert_array_almost_equal(result[i], self.array_m11)

        # Testing with counter=[0,0] and key=[0,0]
        samples = [self.size for x in range(self.nstream)]
        keys = [self.key00 for x in range(self.nstream)]
        counters = [self.counter00 for x in range(self.nstream)]
        result = random_multi(samples, keys, counters, sampler="uniform_m11")
        for i in range(self.nstream):
            self.assertTrue((result[i] > -1.0).all() and (result[i] < 1.0).all())
            np.testing.assert_array_almost_equal(result[i], self.array00_m11)
        return

    def test_rng_01_multi(self):
        # Testing with any counter and any key
        samples = [self.size for x in range(self.nstream)]
        keys = [self.key for x in range(self.nstream)]
        counters = [self.counter for x in range(self.nstream)]
        result = random_multi(samples, keys, counters, sampler="uniform_01")
        for i in range(self.nstream):
            self.assertTrue((result[i] > 0.0).all() and (result[i] < 1.0).all())
            np.testing.assert_array_almost_equal(result[i], self.array_01)

        # Testing with counter=[0,0] and key=[0,0]
        samples = [self.size for x in range(self.nstream)]
        keys = [self.key00 for x in range(self.nstream)]
        counters = [self.counter00 for x in range(self.nstream)]
        result = random_multi(samples, keys, counters, sampler="uniform_01")
        for i in range(self.nstream):
            self.assertTrue((result[i] > 0.0).all() and (result[i] < 1.0).all())
            np.testing.assert_array_almost_equal(result[i], self.array00_01)
        return

    def test_rng_uint64_multi(self):
        # Testing with any counter and any key
        samples = [self.size for x in range(self.nstream)]
        keys = [self.key for x in range(self.nstream)]
        counters = [self.counter for x in range(self.nstream)]
        result = random_multi(samples, keys, counters, sampler="uniform_uint64")
        for i in range(self.nstream):
            self.assertTrue(type(result[i]) == AlignedU64)
            np.testing.assert_array_equal(result[i], self.array_uint64)

        # Testing with counter=[0,0] and key=[0,0]
        samples = [self.size for x in range(self.nstream)]
        keys = [self.key00 for x in range(self.nstream)]
        counters = [self.counter00 for x in range(self.nstream)]
        result = random_multi(samples, keys, counters, sampler="uniform_uint64")
        for i in range(self.nstream):
            self.assertTrue(type(result[i]) == AlignedU64)
            np.testing.assert_array_equal(result[i], self.array00_uint64)
        return



# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

import numpy as np

from .noise import Noise


class AnalyticNoise(Noise):
    """Class representing an analytic noise model.

    This generates an analytic PSD for a set of detectors, given
    input values for the knee frequency, NET, exponent, sample rate,
    minimum frequency, etc.

    Args:
        detectors (list): List of detectors.
        rate (dict): Dictionary of sample rates in Hertz.
        fmin (dict): Dictionary of minimum frequencies for high pass
        fknee (dict): Dictionary of knee frequencies.
        alpha (dict): Dictionary of alpha exponents (positive, not negative!).
        NET (dict): Dictionary of detector NETs.

    """

    def __init__(self, *, detectors, rate, fmin, fknee, alpha, NET, indices=None):

        self._rate = rate
        self._fmin = fmin
        self._fknee = fknee
        self._alpha = alpha
        self._NET = NET

        for d in detectors:
            if self._alpha[d] < 0.0:
                raise RuntimeError(
                    "alpha exponents should be positive in this formalism"
                )

        freqs = {}
        psds = {}

        last_nyquist = None

        for d in detectors:
            if (self._fknee[d] > 0.0) and (self._fknee[d] < self._fmin[d]):
                raise RuntimeError(
                    "If knee frequency is non-zero, it must be greater than f_min"
                )

            nyquist = self._rate[d] / 2.0
            if nyquist != last_nyquist:
                tempfreq = []

                # this starting point corresponds to a high-pass of
                # 30 years, so should be low enough for any interpolation!
                cur = 1.0e-9

                # this value seems to provide a good density of points
                # in log space.
                while cur < nyquist:
                    tempfreq.append(cur)
                    cur *= 1.4

                # put a final point at Nyquist
                tempfreq.append(nyquist)
                tempfreq = np.array(tempfreq, dtype=np.float64)
                last_nyquist = nyquist

            freqs[d] = tempfreq

            if self._fknee[d] > 0.0:
                ktemp = np.power(self._fknee[d], self._alpha[d])
                mtemp = np.power(self._fmin[d], self._alpha[d])
                temp = np.power(freqs[d], self._alpha[d])
                psds[d] = (temp + ktemp) / (temp + mtemp)
                psds[d] *= self._NET[d] * self._NET[d]
            else:
                psds[d] = np.ones_like(freqs[d])
                psds[d] *= self._NET[d] * self._NET[d]

        # call the parent class constructor to store the psds
        super().__init__(detectors=detectors, freqs=freqs, psds=psds, indices=indices)

    def rate(self, det):
        """(float): the sample rate in Hz."""
        return self._rate[det]

    def fmin(self, det):
        """(float): the minimum frequency in Hz, used as a high pass."""
        return self._fmin[det]

    def fknee(self, det):
        """(float): the knee frequency in Hz."""
        return self._fknee[det]

    def alpha(self, det):
        """(float): the (positive!) slope exponent."""
        return self._alpha[det]

    def NET(self, det):
        """(float): the NET."""
        return self._NET[det]



# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

import os

import numpy as np

from ..utils import Logger

from ..timing import function_timer, Timer
from ..rng import random

from ..op import Operator

from toast.mpi import MPI

import toast.qarray as qa

import healpy as hp


class OpSimScanSynchronousSignal(Operator):
    """Operator which generates scan-synchronous signal timestreams.

    Args:
        out (str): accumulate data to the cache with name
            <out>_<detector>.  If the named cache objects do not exist,
            then they are created.
        realization (int): if simulating multiple realizations, the
            realization index.
        component (int): the component index to use for this noise
            simulation.
        nside (int): ground map healpix resolution
        fwhm (int): ground map smoothing scale
        lmax (int): ground map expansion order
        scale (float): RMS of the ground signal fluctuations at el=45deg
        power (float): exponential for suppressing ground pickup at
             higher observing elevation
        path (string): path to a horizontal Healpix map to
            sample for the SSS *instead* of synthesizing Gaussian maps
        report_timing (bool):  Print out time taken to initialize,
             simulate and observe
    """

    def __init__(
        self,
        out="sss",
        realization=0,
        component=663056,
        nside=128,
        fwhm=10,
        lmax=256,
        scale=1e-3,
        power=-1,
        path=None,
        report_timing=False,
    ):
        # Call the parent class constructor
        super().__init__()

        self._out = out
        self._realization = realization
        self._component = component
        self._nside = nside
        self._lmax = lmax
        self._fwhm = fwhm
        self._scale = scale
        self._power = power
        self._path = path
        self._report_timing = report_timing
        return

    @function_timer
    def exec(self, data):
        """Generate timestreams.

        This iterates over all observations and detectors and generates
        the scan-synchronous signal timestreams.

        Args:
            data (toast.Data): The distributed data.

        Returns:
            None

        """

        log = Logger.get()
        group = data.comm.group
        for obs in data.obs:
            try:
                obsname = obs["name"]
            except Exception:
                obsname = "observation"
            prefix = "{} : {} : ".format(group, obsname)
            tod = self._get_from_obs("tod", obs)
            comm = tod.mpicomm
            rank = 0
            if comm is not None:
                rank = comm.rank
            site = self._get_from_obs("site_id", obs)
            weather = self._get_from_obs("weather", obs)

            # Get the observation time span and initialize the weather
            # object if one is provided.
            times = tod.local_times()
            tmin = times[0]
            tmax = times[-1]
            tmin_tot = tmin
            tmax_tot = tmax
            if comm is not None:
                tmin_tot = comm.allreduce(tmin, op=MPI.MIN)
                tmax_tot = comm.allreduce(tmax, op=MPI.MAX)
            weather.set(site, self._realization, tmin_tot)

            key1, key2, counter1, counter2 = self._get_rng_keys(obs)

            if comm is not None:
                comm.Barrier()
            if rank == 0:
                log.info("{}Setting up SSS simulation".format(prefix))

            timer = Timer()
            if self._report_timing:
                if comm is not None:
                    comm.Barrier()
                timer.start()

            sssmap = self._simulate_sss(key1, key2, counter1, counter2, weather, comm)

            self._observe_sss(sssmap, tod, comm, prefix)

            del sssmap

        if self._report_timing:
            if comm is not None:
                comm.Barrier()
            if rank == 0:
                timer.stop()
                timer.report(
                    "{}Simulated and observed scan-synchronous signal" "".format(prefix)
                )
        return

    def _get_from_obs(self, name, obs):
        """Extract value for name from observation.

        If name is not defined in observation, raise an exception.

        """
        if name not in obs:
            raise RuntimeError(
                "Error simulating SSS: observation " 'does not define "{}"'.format(name)
            )
        return obs[name]

    def _get_rng_keys(self, obs):
        """
        The random number generator accepts a key and a counter,
        each made of two 64bit integers.
        Following tod_math.py we set
        key1 = realization * 2^32 + telescope * 2^16 + component
        key2 = obsindx * 2^32
        counter1 = hierarchical cone counter
        counter2 = sample in stream (incremented internally in the atm code)
        """
        telescope = self._get_from_obs("telescope_id", obs)
        site = self._get_from_obs("site_id", obs)
        obsindx = self._get_from_obs("id", obs)
        key1 = self._realization * 2 ** 32 + telescope * 2 ** 16 + self._component
        key2 = site * 2 ** 16 + obsindx
        counter1 = 0
        counter2 = 0
        return key1, key2, counter1, counter2

    def _simulate_sss(self, key1, key2, counter1, counter2, weather, comm):
        """
        Create a map of the ground signal to observe with all detectors
        """
        # FIXME: we could store the map in node-shared memory
        #
        # Surface temperature is made available but not used yet
        # to scale the SSS
        if comm is None or comm.rank == 0:
            # Only the root process loads or simulates the map
            temperature = weather.surface_temperature
            if self._path:
                sssmap = hp.read_map(self._path, verbose=False)
            else:
                npix = 12 * self._nside ** 2
                sssmap = random(
                    npix,
                    key=(key1, key2),
                    counter=(counter1, counter2),
                    sampler="gaussian",
                )
                sssmap = np.array(sssmap, dtype=np.float)
                sssmap = hp.smoothing(
                    sssmap, fwhm=np.radians(self._fwhm), lmax=self._lmax
                )
                sssmap /= np.std(sssmap)
                lon, lat = hp.pix2ang(
                    self._nside, np.arange(npix, dtype=np.int), lonlat=True
                )
                scale = self._scale * (np.abs(lat) / 90 + 0.5) ** self._power
                sssmap *= scale
        else:
            sssmap = None
        if comm is not None:
            sssmap = comm.bcast(sssmap)
        return sssmap

    @function_timer
    def _observe_sss(self, sssmap, tod, comm, prefix):
        """
        Use healpy bilinear interpolation to observe the ground signal map
        """
        log = Logger.get()
        rank = 0
        if comm is not None:
            rank = comm.rank
        timer = Timer()
        if self._report_timing:
            if comm is not None:
                comm.Barrier()
            timer.start()

        nsamp = tod.local_samples[1]

        if rank == 0:
            log.info("{}Observing the scan-synchronous signal".format(prefix))

        for det in tod.local_dets:
            # Cache the output signal
            cachename = "{}_{}".format(self._out, det)
            if tod.cache.exists(cachename):
                ref = tod.cache.reference(cachename)
            else:
                ref = tod.cache.create(cachename, np.float64, (nsamp,))

            try:
                # Some TOD classes provide a shortcut to Az/El
                az, el = tod.read_azel(detector=det)
                phi = 2 * np.pi - az
                theta = np.pi / 2 - el
            except Exception as e:
                azelquat = tod.read_pntg(detector=det, azel=True)
                # Convert Az/El quaternion of the detector back into
                # angles for the simulation.
                theta, phi = qa.to_position(azelquat)
                # Azimuth is measured in the opposite direction
                # than longitude
                # az = 2 * np.pi - phi
                # el = np.pi / 2 - theta

            ref[:] += hp.get_interp_val(sssmap, theta, phi)

            del ref

        if self._report_timing:
            if comm is not None:
                comm.Barrier()
            if rank == 0:
                timer.stop()
                timer.report("{}OpSimSSS: Observe signal".format(prefix))
        return



# Copyright (c) 2019-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

import argparse
import os

import numpy as np

from ..timing import function_timer, Timer
from ..utils import Logger, Environment

from ..todmap import OpSimScanSynchronousSignal


def add_sss_args(parser):
    """Add the sky arguments"""

    parser.add_argument(
        "--ground-map", required=False, help="Fixed ground template map"
    )
    parser.add_argument(
        "--ground-nside",
        required=False,
        default=128,
        type=np.int,
        help="Ground template resolution",
    )
    parser.add_argument(
        "--ground-fwhm-deg",
        required=False,
        default=10,
        type=np.float,
        help="Ground template smoothing in degrees",
    )
    parser.add_argument(
        "--ground-lmax",
        required=False,
        default=256,
        type=np.int,
        help="Ground template expansion order",
    )
    parser.add_argument(
        "--ground-scale",
        required=False,
        default=1e-3,
        type=np.float,
        help="Ground template RMS at el=45 deg",
    )
    parser.add_argument(
        "--ground-power",
        required=False,
        default=-1,
        type=np.float,
        help="Exponential for suppressing ground pick-up at "
        "higher observing elevation",
    )

    parser.add_argument(
        "--simulate-ground",
        required=False,
        action="store_true",
        help="Enable simulating ground pickup.",
        dest="simulate_ground",
    )
    parser.add_argument(
        "--no-simulate-ground",
        required=False,
        action="store_false",
        help="Disable simulating ground pickup.",
        dest="simulate_ground",
    )
    parser.set_defaults(simulate_ground=False)

    return


@function_timer
def simulate_sss(args, comm, data, mc, cache_prefix=None, verbose=False):
    if not args.simulate_ground:
        return
    log = Logger.get()
    timer = Timer()
    if comm.world_rank == 0 and verbose:
        log.info("Simulating sss")
    timer.start()
    nse = OpSimScanSynchronousSignal(
        out=cache_prefix,
        realization=mc,
        nside=args.ground_nside,
        fwhm=args.ground_fwhm_deg,
        scale=args.ground_scale,
        lmax=args.ground_lmax,
        power=args.ground_power,
        path=args.ground_map,
        report_timing=verbose,
    )
    nse.exec(data)
    if comm.comm_world is not None:
        comm.comm_world.barrier()
    timer.stop()
    if comm.world_rank == 0 and verbose:
        timer.report("sss simulation")
    return



# Copyright (c) 2019-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

import argparse
import os

import numpy as np

from ..timing import function_timer, Timer
from ..todmap import OpSimDipole
from ..utils import Logger, Environment


def add_dipole_args(parser):
    parser.add_argument(
        "--dipole",
        required=False,
        action="store_true",
        help="Add simulated dipole",
        dest="simulate_dipole",
    )
    parser.add_argument(
        "--no-dipole",
        required=False,
        action="store_false",
        help="Do not add simulated dipole",
        dest="simulate_dipole",
    )
    parser.set_defaults(simulate_dipole=False)
    parser.add_argument(
        "--dipole-mode",
        required=False,
        default="total",
        help="Dipole mode is 'total', 'orbital' or 'solar'",
    )
    parser.add_argument(
        "--dipole-solar-speed-kms",
        required=False,
        help="Solar system speed [km/s]",
        type=float,
        default=369.0,
    )
    parser.add_argument(
        "--dipole-solar-gal-lat-deg",
        required=False,
        help="Solar system speed galactic latitude [degrees]",
        type=float,
        default=48.26,
    )
    parser.add_argument(
        "--dipole-solar-gal-lon-deg",
        required=False,
        help="Solar system speed galactic longitude[degrees]",
        type=float,
        default=263.99,
    )

    # `coord` may already be added
    try:
        parser.add_argument(
            "--coord", required=False, default="C", help="Sky coordinate system [C,E,G]"
        )
    except argparse.ArgumentError:
        pass
    return


def simulate_dipole(args, comm, data, cache_prefix, freq=0, verbose=True):
    """Simulate CMB dipole and add it to `cache_prefix`

    Args:
        freq (float) :  Observing frequency in GHz.  If non-zero, the
            relativistic quadrupole will include frequency corrections.
    """
    if not args.simulate_dipole:
        return None
    log = Logger.get()
    timer = Timer()
    timer.start()
    if comm.world_rank == 0 and verbose:
        log.info("Simulating dipole")
    has_signal = True
    op_sim_dipole = OpSimDipole(
        mode=args.dipole_mode,
        solar_speed=args.dipole_solar_speed_kms,
        solar_gal_lat=args.dipole_solar_gal_lat_deg,
        solar_gal_lon=args.dipole_solar_gal_lon_deg,
        out=cache_prefix,
        keep_quats=False,
        keep_vel=False,
        subtract=False,
        coord=args.coord,
        freq=freq,
        flag_mask=255,
        common_flag_mask=255,
    )
    op_sim_dipole.exec(data)
    if comm.world_rank == 0 and verbose:
        timer.report_clear("Simulate dipole")
    return cache_prefix



# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

from .mpi import MPITestCase

import os
import shutil

import numpy as np
import numpy.testing as nt

import healpy as hp

from ..tod import AnalyticNoise, OpSimNoise
from ..todmap import TODGround, OpPointingHpix, OpSimGradient, OpSimScan, OpMadam

from ..map import DistPixels

from ._helpers import create_outdir, create_distdata, boresight_focalplane


class MapGroundTest(MPITestCase):
    def setUp(self):
        fixture_name = os.path.splitext(os.path.basename(__file__))[0]
        self.outdir = create_outdir(self.comm, fixture_name)

        # Create one observation per group, and each observation will have
        # one detector per process and a single chunk.

        self.data = create_distdata(self.comm, obs_per_group=1)
        self.data_fast_hwp = create_distdata(self.comm, obs_per_group=1)
        self.data_const_hwp = create_distdata(self.comm, obs_per_group=1)

        self.ndet = self.data.comm.group_size
        self.rate = 20.0

        # Create detectors with white noise
        self.NET = 5.0

        (
            dnames,
            dquat,
            depsilon,
            drate,
            dnet,
            dfmin,
            dfknee,
            dalpha,
        ) = boresight_focalplane(
            self.ndet, samplerate=self.rate, fknee=0.0, net=self.NET
        )

        # Samples per observation
        self.totsamp = 100000

        # Pixelization
        nside = 256
        self.sim_nside = nside
        self.map_nside = nside

        # Scan properties
        self.site_lon = "-67:47:10"
        self.site_lat = "-22:57:30"
        self.site_alt = 5200.0
        self.coord = "C"
        self.azmin = 45
        self.azmax = 55
        self.el = 60
        self.scanrate = 1.0
        self.scan_accel = 0.1
        self.CES_start = None

        # Populate the single observation per group

        common_args = [dquat, self.totsamp]
        common_kwargs = {
            "detranks": self.data.comm.group_size,
            "firsttime": 0.0,
            "rate": self.rate,
            "site_lon": self.site_lon,
            "site_lat": self.site_lat,
            "site_alt": self.site_alt,
            "azmin": self.azmin,
            "azmax": self.azmax,
            "el": self.el,
            "coord": self.coord,
            "scanrate": self.scanrate,
            "scan_accel": self.scan_accel,
            "CES_start": self.CES_start,
        }

        # No HWP
        tod_no_hwp = TODGround(self.data.comm.comm_group, *common_args, **common_kwargs)

        # CHWP
        hwprate = self.rate * 60.0
        tod_fast_hwp = TODGround(
            self.data_fast_hwp.comm.comm_group,
            *common_args,
            **common_kwargs,
            hwprpm=hwprate
        )

        # Stepped HWP
        hwpstep = 2.0 * np.pi
        hwpsteptime = (self.totsamp / self.rate) / 60.0
        tod_const_hwp = TODGround(
            self.data_const_hwp.comm.comm_group,
            *common_args,
            **common_kwargs,
            hwpstep=hwpstep,
            hwpsteptime=hwpsteptime
        )

        for data, tod in [
            (self.data, tod_no_hwp),
            (self.data_fast_hwp, tod_fast_hwp),
            (self.data_const_hwp, tod_const_hwp),
        ]:

            self.common_flag_mask = tod.TURNAROUND

            common_flags = tod.read_common_flags()

            # Number of flagged samples in each observation.  Only the first row
            # of the process grid needs to contribute, since all process columns
            # have identical common flags.
            nflagged = 0
            if (tod.grid_comm_col is None) or (tod.grid_comm_col.rank == 0):
                nflagged += np.sum((common_flags & self.common_flag_mask) != 0)

            # Number of flagged samples across all observations
            self.nflagged = None
            if self.comm is None:
                self.nflagged = nflagged
            else:
                self.nflagged = self.data.comm.comm_world.allreduce(nflagged)

            # add analytic noise model with white noise

            nse = AnalyticNoise(
                rate=drate,
                fmin=dfmin,
                detectors=dnames,
                fknee=dfknee,
                alpha=dalpha,
                NET=dnet,
            )

            data.obs[0]["tod"] = tod
            data.obs[0]["noise"] = nse
        return

    def test_azel(self):
        quats1 = []
        quats2 = []

        for ob in self.data.obs:
            tod = ob["tod"]
            for d in tod.local_dets:
                quats1.append(tod.read_pntg(detector=d))
                quats2.append(tod.read_pntg(detector=d, azel=True))

        for i in range(10):
            if np.all(quats1[0][i] == quats2[0][i]):
                raise Exception("Horizontal and celestial pointing must be different")
        return

    def test_grad(self):
        rank = 0
        if self.comm is not None:
            rank = self.comm.rank
        # add simple sky gradient signal
        grad = OpSimGradient(
            nside=self.sim_nside, nest=True, common_flag_mask=self.common_flag_mask
        )
        grad.exec(self.data)

        # make a simple pointing matrix
        pointing = OpPointingHpix(nside=self.map_nside, nest=True)
        pointing.exec(self.data)

        handle = None
        if rank == 0:
            handle = open(os.path.join(self.outdir, "out_test_ground_grad_info"), "w")
        self.data.info(handle=handle, common_flag_mask=self.common_flag_mask)
        if rank == 0:
            handle.close()

        # make a binned map with madam
        madam_out = os.path.join(self.outdir, "madam_grad")
        if rank == 0:
            if os.path.isdir(madam_out):
                shutil.rmtree(madam_out)
            os.mkdir(madam_out)

        pars = {}
        pars["kfirst"] = "F"
        pars["base_first"] = 1.0
        pars["fsample"] = self.rate
        pars["nside_map"] = self.map_nside
        pars["nside_cross"] = self.map_nside
        pars["nside_submap"] = min(8, self.map_nside)
        pars["write_map"] = "F"
        pars["write_binmap"] = "T"
        pars["write_matrix"] = "F"
        pars["write_wcov"] = "F"
        pars["write_hits"] = "T"
        pars["kfilter"] = "F"
        pars["path_output"] = madam_out
        pars["info"] = 0

        madam = OpMadam(
            params=pars,
            name="grad",
            purge=False,
            common_flag_mask=self.common_flag_mask,
        )
        if madam.available:
            madam.exec(self.data)

            if rank == 0:
                import matplotlib.pyplot as plt

                hitsfile = os.path.join(madam_out, "madam_hmap.fits")
                hits = hp.read_map(hitsfile, nest=True)

                outfile = "{}.png".format(hitsfile)
                hp.mollview(hits, xsize=1600, nest=True)
                plt.savefig(outfile)
                plt.close()

                binfile = os.path.join(madam_out, "madam_bmap.fits")
                bins = hp.read_map(binfile, nest=True)

                outfile = "{}.png".format(binfile)
                hp.mollview(bins, xsize=1600, nest=True)
                plt.savefig(outfile)
                plt.close()

                # compare binned map to input signal

                tothits = np.sum(hits)
                nt.assert_equal(
                    self.ndet
                    * ((self.data.comm.ngroups * self.totsamp) - self.nflagged),
                    tothits,
                )

                sig = grad.sigmap()
                mask = bins > -1.0e20
                nt.assert_almost_equal(bins[mask], sig[mask], decimal=4)
        else:
            print("libmadam not available, skipping tests")
        return

    def test_noise(self):
        rank = 0
        if self.comm is not None:
            rank = self.comm.rank
        # generate noise timestreams from the noise model
        nsig = OpSimNoise()
        nsig.exec(self.data)

        # make a simple pointing matrix
        pointing = OpPointingHpix(nside=self.map_nside, nest=True)
        pointing.exec(self.data)

        handle = None
        if rank == 0:
            handle = open(os.path.join(self.outdir, "out_test_ground_noise_info"), "w")
        self.data.info(handle=handle, common_flag_mask=self.common_flag_mask)
        if rank == 0:
            handle.close()

        # For noise weighting in madam, we know we are using an analytic noise
        # and so we can use noise weights based on the NET.  This is instrument
        # specific.

        tod = self.data.obs[0]["tod"]
        nse = self.data.obs[0]["noise"]
        detweights = {}
        for d in tod.local_dets:
            detweights[d] = 1.0 / (self.rate * nse.NET(d) ** 2)

        # make a binned map with madam
        madam_out = os.path.join(self.outdir, "madam_noise")
        if rank == 0:
            if os.path.isdir(madam_out):
                shutil.rmtree(madam_out)
            os.mkdir(madam_out)

        pars = {}
        pars["kfirst"] = "F"
        pars["base_first"] = 1.0
        pars["fsample"] = self.rate
        pars["nside_map"] = self.map_nside
        pars["nside_cross"] = self.map_nside
        pars["nside_submap"] = min(8, self.map_nside)
        pars["write_map"] = "F"
        pars["write_binmap"] = "T"
        pars["write_matrix"] = "F"
        pars["write_wcov"] = "F"
        pars["write_hits"] = "T"
        pars["kfilter"] = "F"
        pars["path_output"] = madam_out
        pars["info"] = 0

        madam = OpMadam(
            params=pars,
            detweights=detweights,
            name="noise",
            common_flag_mask=self.common_flag_mask,
        )
        if madam.available:
            madam.exec(self.data)

            if rank == 0:
                import matplotlib.pyplot as plt

                hitsfile = os.path.join(madam_out, "madam_hmap.fits")
                hits = hp.read_map(hitsfile, nest=True)

                outfile = "{}.png".format(hitsfile)
                hp.mollview(hits, xsize=1600, nest=True)
                plt.savefig(outfile)
                plt.close()

                binfile = os.path.join(madam_out, "madam_bmap.fits")
                bins = hp.read_map(binfile, nest=True)

                outfile = "{}.png".format(binfile)
                hp.mollview(bins, xsize=1600, nest=True)
                plt.savefig(outfile)
                plt.close()

                # check that pixel rms makes sense given the
                # number of hits and the timestream rms

                tothits = np.sum(hits)
                nt.assert_equal(
                    self.ndet
                    * ((self.data.comm.ngroups * self.totsamp) - self.nflagged),
                    tothits,
                )

                mask = bins > -1.0e20
                # print("num good pix = ", len(mask))
                rthits = np.sqrt(hits[mask].astype(np.float64))
                # print("rthits = ", rthits)
                # print("bmap = ", bins[mask])
                weighted = bins[mask] * rthits
                # print("weighted = ", weighted)
                pixrms = np.sqrt(np.mean(weighted ** 2))
                todrms = self.NET * np.sqrt(self.rate)
                relerr = np.absolute(pixrms - todrms) / todrms
                # print("pixrms = ", pixrms)
                # print("todrms = ", todrms)
                # print("relerr = ", relerr)
                self.assertTrue(relerr < 0.03)
        else:
            print("libmadam not available, skipping tests")
        return

    def test_scanmap(self):
        rank = 0
        if self.comm is not None:
            rank = self.comm.rank
        # make a simple pointing matrix
        pointing = OpPointingHpix(nside=self.map_nside, nest=True)
        pointing.exec(self.data)

        # construct a sky gradient operator, just to get the signal
        # map- we are not going to use the operator on the data.
        grad = OpSimGradient(
            nside=self.sim_nside, nest=True, common_flag_mask=self.common_flag_mask
        )
        sig = grad.sigmap()

        # construct a distributed map which has the gradient
        distsig = DistPixels(
            self.data, comm=self.data.comm.comm_group, nnz=1, dtype=np.float64
        )
        distsig.broadcast_healpix_map(sig)

        # create TOD from map
        scansim = OpSimScan(input_map=distsig)
        scansim.exec(self.data)

        handle = None
        if rank == 0:
            handle = open(
                os.path.join(self.outdir, "out_test_ground_scanmap_info"), "w"
            )
        self.data.info(handle=handle, common_flag_mask=self.common_flag_mask)
        if rank == 0:
            handle.close()

        # make a binned map with madam
        madam_out = os.path.join(self.outdir, "madam_scansim")
        if rank == 0:
            if os.path.isdir(madam_out):
                shutil.rmtree(madam_out)
            os.mkdir(madam_out)

        pars = {}
        pars["kfirst"] = "F"
        pars["base_first"] = 1.0
        pars["fsample"] = self.rate
        pars["nside_map"] = self.map_nside
        pars["nside_cross"] = self.map_nside
        pars["nside_submap"] = min(8, self.map_nside)
        pars["write_map"] = "F"
        pars["write_binmap"] = "T"
        pars["write_matrix"] = "F"
        pars["write_wcov"] = "F"
        pars["write_hits"] = "T"
        pars["kfilter"] = "F"
        pars["path_output"] = madam_out
        pars["info"] = 0

        madam = OpMadam(
            params=pars, name="scan", common_flag_mask=self.common_flag_mask
        )
        if madam.available:
            madam.exec(self.data)

            if rank == 0:
                import matplotlib.pyplot as plt

                hitsfile = os.path.join(madam_out, "madam_hmap.fits")
                hits = hp.read_map(hitsfile, nest=True)

                outfile = "{}.png".format(hitsfile)
                hp.mollview(hits, xsize=1600, nest=True)
                plt.savefig(outfile)
                plt.close()

                binfile = os.path.join(madam_out, "madam_bmap.fits")
                bins = hp.read_map(binfile, nest=True)

                outfile = "{}.png".format(binfile)
                hp.mollview(bins, xsize=1600, nest=True)
                plt.savefig(outfile)
                plt.close()

                # compare binned map to input signal

                tothits = np.sum(hits)
                nt.assert_equal(
                    self.ndet
                    * ((self.data.comm.ngroups * self.totsamp) - self.nflagged),
                    tothits,
                )
                mask = bins > -1.0e20
                nt.assert_almost_equal(bins[mask], sig[mask], decimal=4)
        else:
            print("libmadam not available, skipping tests")
        return

    def test_hwpfast(self):
        rank = 0
        if self.comm is not None:
            rank = self.comm.rank
        # make a pointing matrix with a HWP that rotates 2*PI every sample
        data = self.data_fast_hwp
        pointing = OpPointingHpix(nside=self.map_nside, nest=True)
        pointing.exec(data)

        # construct a sky gradient operator, just to get the signal
        # map- we are not going to use the operator on the data.
        grad = OpSimGradient(nside=self.sim_nside, nest=True)
        sig = grad.sigmap()

        # construct a distributed map which has the gradient
        distsig = DistPixels(data, comm=data.comm.comm_group, nnz=1, dtype=np.float64)
        distsig.broadcast_healpix_map(sig)

        # create TOD from map
        scansim = OpSimScan(input_map=distsig)
        scansim.exec(data)

        handle = None
        if rank == 0:
            handle = open(
                os.path.join(self.outdir, "out_test_ground_hwpfast_info"), "w"
            )
        data.info(handle=handle, common_flag_mask=self.common_flag_mask)
        if rank == 0:
            handle.close()

        # make a binned map with madam
        madam_out = os.path.join(self.outdir, "madam_hwpfast")
        if rank == 0:
            if os.path.isdir(madam_out):
                shutil.rmtree(madam_out)
            os.mkdir(madam_out)

        pars = {}
        pars["kfirst"] = "F"
        pars["base_first"] = 1.0
        pars["fsample"] = self.rate
        pars["nside_map"] = self.map_nside
        pars["nside_cross"] = self.map_nside
        pars["nside_submap"] = min(8, self.map_nside)
        pars["write_map"] = "F"
        pars["write_binmap"] = "T"
        pars["write_matrix"] = "F"
        pars["write_wcov"] = "F"
        pars["write_hits"] = "T"
        pars["kfilter"] = "F"
        pars["path_output"] = madam_out
        pars["info"] = 0

        madam = OpMadam(
            params=pars, name="scan", common_flag_mask=self.common_flag_mask
        )
        if madam.available:
            madam.exec(data)

            if rank == 0:
                import matplotlib.pyplot as plt

                hitsfile = os.path.join(madam_out, "madam_hmap.fits")
                hits = hp.read_map(hitsfile, nest=True)

                outfile = "{}.png".format(hitsfile)
                hp.mollview(hits, xsize=1600, nest=True)
                plt.savefig(outfile)
                plt.close()

                binfile = os.path.join(madam_out, "madam_bmap.fits")
                bins = hp.read_map(binfile, nest=True)

                outfile = "{}.png".format(binfile)
                hp.mollview(bins, xsize=1600, nest=True)
                plt.savefig(outfile)
                plt.close()

                # compare binned map to input signal

                tothits = np.sum(hits)
                nt.assert_equal(
                    self.ndet * ((data.comm.ngroups * self.totsamp) - self.nflagged),
                    tothits,
                )
                mask = bins > -1.0e20
                nt.assert_almost_equal(bins[mask], sig[mask], decimal=4)
        else:
            print("libmadam not available, skipping tests")
        return

    def test_hwpconst(self):
        rank = 0
        if self.comm is not None:
            rank = self.comm.rank
        # make a pointing matrix with a HWP that is constant
        data = self.data_const_hwp
        pointing = OpPointingHpix(nside=self.map_nside, nest=True)
        pointing.exec(data)

        # construct a sky gradient operator, just to get the signal
        # map- we are not going to use the operator on the data.
        grad = OpSimGradient(nside=self.sim_nside, nest=True)
        sig = grad.sigmap()

        # construct a distributed map which has the gradient
        distsig = DistPixels(data, comm=data.comm.comm_group, nnz=1, dtype=np.float64)
        distsig.broadcast_healpix_map(sig)

        # create TOD from map
        scansim = OpSimScan(input_map=distsig)
        scansim.exec(data)

        handle = None
        if rank == 0:
            handle = open(
                os.path.join(self.outdir, "out_test_ground_hwpconst_info"), "w"
            )
        data.info(handle=handle, common_flag_mask=self.common_flag_mask)
        if rank == 0:
            handle.close()

        # make a binned map with madam
        madam_out = os.path.join(self.outdir, "madam_hwpconst")
        if rank == 0:
            if os.path.isdir(madam_out):
                shutil.rmtree(madam_out)
            os.mkdir(madam_out)

        pars = {}
        pars["kfirst"] = "F"
        pars["base_first"] = 1.0
        pars["fsample"] = self.rate
        pars["nside_map"] = self.map_nside
        pars["nside_cross"] = self.map_nside
        pars["nside_submap"] = min(8, self.map_nside)
        pars["write_map"] = "F"
        pars["write_binmap"] = "T"
        pars["write_matrix"] = "F"
        pars["write_wcov"] = "F"
        pars["write_hits"] = "T"
        pars["kfilter"] = "F"
        pars["path_output"] = madam_out
        pars["info"] = 0

        madam = OpMadam(
            params=pars, name="scan", common_flag_mask=self.common_flag_mask
        )
        if madam.available:
            madam.exec(data)

            if rank == 0:
                import matplotlib.pyplot as plt

                hitsfile = os.path.join(madam_out, "madam_hmap.fits")
                hits = hp.read_map(hitsfile, nest=True)

                outfile = "{}.png".format(hitsfile)
                hp.mollview(hits, xsize=1600, nest=True)
                plt.savefig(outfile)
                plt.close()

                binfile = os.path.join(madam_out, "madam_bmap.fits")
                bins = hp.read_map(binfile, nest=True)

                outfile = "{}.png".format(binfile)
                hp.mollview(bins, xsize=1600, nest=True)
                plt.savefig(outfile)
                plt.close()

                # compare binned map to input signal

                tothits = np.sum(hits)
                nt.assert_equal(
                    self.ndet * ((data.comm.ngroups * self.totsamp) - self.nflagged),
                    tothits,
                )
                mask = bins > -1.0e20
                nt.assert_almost_equal(bins[mask], sig[mask], decimal=4)
        else:
            print("libmadam not available, skipping tests")
        return



# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

import os

import numpy as np

import healpy as hp

from .mpi import MPITestCase

from ..mpi import MPI

from .. import healpix as hpx

from .. import qarray as qa

from ..todmap import slew_precession_axis, satellite_scanning, TODSatellite

from ._helpers import create_outdir, create_distdata, boresight_focalplane


class TODSatelliteTest(MPITestCase):
    def setUp(self):
        fixture_name = os.path.splitext(os.path.basename(__file__))[0]
        self.outdir = create_outdir(self.comm, fixture_name)

        # Create 366 observations, divided evenly between groups
        self.nobs = 366
        opg = self.nobs
        if (self.comm is not None) and (self.comm.size >= 2):
            opg = self.nobs // 2
        self.data = create_distdata(self.comm, obs_per_group=opg)

        # One boresight detector
        self.ndet = 1
        self.rate = 1.0 / 60.0

        (
            dnames,
            dquat,
            depsilon,
            drate,
            dnet,
            dfmin,
            dfknee,
            dalpha,
        ) = boresight_focalplane(self.ndet, samplerate=self.rate)

        # Scan strategy.
        # Choose scan parameters so that we return to the origin.
        precangle = 35.0
        spinangle = 55.0
        spinperiod = 240
        precperiod = 8640

        # Precession axis slew
        degday = 360.0 / self.nobs

        # Sampling

        # Only simulate 22 out of 24 hours per observation, in order to test
        # that the starting phase is propagated.  Put the "gap" at the start
        # of each day, so that we can look at the final data point of the
        # final observation.

        daysamps = 24 * 60
        # goodsamps = 24 * 60
        goodsamps = 22 * 60
        badsamps = daysamps - goodsamps

        # Populate the observations
        for oid, obs in enumerate(self.data.obs):
            firstsamp = oid * daysamps + badsamps

            # On the last observation, simulate one extra sample so we get
            # back to the starting point.
            nsim = goodsamps
            if (oid == len(self.data.obs) - 1) and (
                self.data.comm.group == self.data.comm.ngroups - 1
            ):
                nsim += 1

            tod = TODSatellite(
                self.data.comm.comm_group,
                dquat,
                nsim,
                detranks=1,
                firstsamp=firstsamp,
                firsttime=(firstsamp / self.rate),
                rate=self.rate,
                spinperiod=spinperiod,
                spinangle=spinangle,
                precperiod=precperiod,
                precangle=precangle,
            )

            qprec = np.empty(4 * tod.local_samples[1], dtype=np.float64).reshape(
                (-1, 4)
            )

            slew_precession_axis(
                qprec,
                firstsamp=(firstsamp + tod.local_samples[0]),
                samplerate=self.rate,
                degday=degday,
            )

            tod.set_prec_axis(qprec=qprec)

            obs["tod"] = tod

    def tearDown(self):
        pass

    def test_precession(self):
        # Test precession axis slew
        slewrate = 1.0 / (24.0 * 3600.0)
        nobs = 366
        degday = 360.0 / nobs

        # Plus one is to get back to the beginning
        nsim = nobs + 1

        qprec = np.empty(4 * nsim, dtype=np.float64).reshape((-1, 4))

        slew_precession_axis(qprec, firstsamp=0, samplerate=slewrate, degday=degday)

        zaxis = np.array([0.0, 0.0, 1.0])

        v = qa.rotate(qprec, zaxis)

        dotprod = v[0][0] * v[-1][0] + v[0][1] * v[-1][1] + v[0][2] * v[-1][2]
        np.testing.assert_almost_equal(dotprod, 1.0)
        return

    def test_phase(self):
        # Choose scan parameters so that we return to the origin.
        nobs = 366
        precangle = 35.0
        spinangle = 55.0
        spinperiod = 240
        precperiod = 8640

        # Precession axis slew
        degday = 360.0 / 366.0

        samplerate = 1.0 / 60.0

        daysamps = 24 * 60
        goodsamps = 24 * 60
        # goodsamps = 22 * 60
        badsamps = daysamps - goodsamps

        # hit map
        nside = 32
        pix = hpx.Pixels(nside)
        pdata = np.zeros(12 * nside * nside, dtype=np.int32)

        # Only simulate 22 out of 24 hours per observation, in order to test
        # that the starting phase is propagated.  Put the "gap" at the start
        # of each day, so that we can look at the final data point of the
        # final observation.

        zaxis = np.array([0.0, 0.0, 1.0])

        vlast = None

        for ob in range(nobs):
            firstsamp = ob * daysamps + badsamps

            # On the last observation, simulate one extra sample so we get
            # back to the starting point.
            nsim = goodsamps
            if ob == nobs - 1:
                nsim += 1

            qprec = np.empty(4 * nsim, dtype=np.float64).reshape((-1, 4))
            slew_precession_axis(
                qprec, firstsamp=firstsamp, samplerate=samplerate, degday=degday
            )

            boresight = np.empty(4 * nsim, dtype=np.float64).reshape((-1, 4))
            satellite_scanning(
                boresight,
                firstsamp=firstsamp,
                samplerate=samplerate,
                qprec=qprec,
                spinperiod=spinperiod,
                spinangle=spinangle,
                precperiod=precperiod,
                precangle=precangle,
            )

            v = qa.rotate(boresight, zaxis)
            p = pix.vec2ring(v)
            pdata[p] += 1
            vlast = v[-1]

        if self.data.comm.world_rank == 0:
            import matplotlib.pyplot as plt

            hitsfile = os.path.join(self.outdir, "tod_satellite_hits.fits")
            if os.path.isfile(hitsfile):
                os.remove(hitsfile)
            hp.write_map(hitsfile, pdata, nest=False, dtype=np.int32)
            outfile = "{}.png".format(hitsfile)
            hp.mollview(pdata, xsize=1600, nest=False)
            plt.savefig(outfile)
            plt.close()

        np.testing.assert_almost_equal(vlast[0], 0.0)
        np.testing.assert_almost_equal(vlast[1], -1.0)
        np.testing.assert_almost_equal(vlast[2], 0.0)
        return

    def test_todclass(self):
        # Hit map
        nside = 32
        pix = hpx.Pixels(nside)
        pdata = np.zeros(12 * nside * nside, dtype=np.int32)

        # Compute the boresight hitmap

        zaxis = np.array([0.0, 0.0, 1.0])
        vlast = None

        for obs in self.data.obs:
            tod = obs["tod"]
            boresight = tod.read_boresight()
            v = qa.rotate(boresight, zaxis)
            p = pix.vec2ring(v)
            pdata[p] += 1
            vlast = v[-1]

        allpdata = None
        if self.data.comm.comm_world is None:
            allpdata = pdata
        else:
            if self.data.comm.world_rank == 0:
                allpdata = np.zeros_like(pdata)
            self.data.comm.comm_world.Reduce(pdata, allpdata, op=MPI.SUM, root=0)

        if self.data.comm.world_rank == 0:
            import matplotlib.pyplot as plt

            hitsfile = os.path.join(self.outdir, "tod_satellite_classhits.fits")
            if os.path.isfile(hitsfile):
                os.remove(hitsfile)
            hp.write_map(hitsfile, allpdata, nest=False, dtype=np.int32)

            outfile = "{}.png".format(hitsfile)
            hp.mollview(allpdata, xsize=1600, nest=False)
            plt.savefig(outfile)
            plt.close()

        # The last sample on the last process should be back to the origin
        lastpass = True
        lastproc = self.data.comm.world_size - 1
        if self.data.comm.world_rank == lastproc:
            lastpass = np.allclose(vlast[0], 0.0)
            lastpass = np.allclose(vlast[1], -1.0)
            lastpass = np.allclose(vlast[2], 0.0)
        if self.data.comm.comm_world is not None:
            lastpass = self.data.comm.comm_world.bcast(lastpass, root=lastproc)
        self.assertTrue(lastpass)

        return



# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

import os

import healpy as hp
import numpy as np

from ..map import DistPixels
from ..mpi import MPI
from ..timing import function_timer, GlobalTimers
from ..utils import Logger, Environment

from .. import qarray as qa

from .._libtoast import scan_map_float64, scan_map_float32

from ..op import Operator


class OpSimGradient(Operator):
    """Generate a fake sky signal as a gradient between the poles.

    This passes through each observation and creates a fake signal timestream
    based on the cartesian Z coordinate of the HEALPix pixel containing the
    detector pointing.

    Args:
        out (str): accumulate data to the cache with name <out>_<detector>.
            If the named cache objects do not exist, then they are created.
        nside (int): the HEALPix NSIDE value to use.
        min (float): the minimum value to use at the South Pole.
        max (float): the maximum value to use at the North Pole.
        nest (bool): whether to use NESTED ordering.
    """

    def __init__(
        self,
        out="grad",
        nside=512,
        min=-100.0,
        max=100.0,
        nest=False,
        flag_mask=255,
        common_flag_mask=255,
        keep_quats=False,
    ):
        # Call the parent class constructor
        super().__init__()
        self._nside = nside
        self._out = out
        self._min = min
        self._max = max
        self._nest = nest
        self._flag_mask = flag_mask
        self._common_flag_mask = common_flag_mask
        self._keep_quats = keep_quats

    @function_timer
    def exec(self, data):
        """Create the gradient timestreams.

        This pixelizes each detector's pointing and then assigns a
        timestream value based on the cartesian Z coordinate of the pixel
        center.

        Args:
            data (toast.Data): The distributed data.

        """
        zaxis = np.array([0, 0, 1], dtype=np.float64)
        nullquat = np.array([0, 0, 0, 1], dtype=np.float64)

        range = self._max - self._min

        for obs in data.obs:
            tod = obs["tod"]

            offset, nsamp = tod.local_samples

            common = tod.local_common_flags() & self._common_flag_mask

            for det in tod.local_dets:
                flags = tod.local_flags(det) & self._flag_mask
                totflags = flags | common
                del flags

                pdata = tod.local_pointing(det).copy()
                pdata[totflags != 0, :] = nullquat

                dir = qa.rotate(pdata, zaxis)

                pixels = hp.vec2pix(
                    self._nside, dir[:, 0], dir[:, 1], dir[:, 2], nest=self._nest
                )
                x, y, z = hp.pix2vec(self._nside, pixels, nest=self._nest)
                z += 1.0
                z *= 0.5
                z *= range
                z += self._min
                z[totflags != 0] = 0.0

                cachename = "{}_{}".format(self._out, det)
                if not tod.cache.exists(cachename):
                    tod.cache.create(cachename, np.float64, (nsamp,))
                ref = tod.cache.reference(cachename)
                ref[:] += z
                del ref

                if not self._keep_quats:
                    cachename = "quat_{}".format(det)
                    tod.cache.destroy(cachename)

            del common
        return

    def sigmap(self):
        """(array): Return the underlying signal map (full map on all processes)."""
        range = self._max - self._min
        pix = np.arange(0, 12 * self._nside * self._nside, dtype=np.int64)
        x, y, z = hp.pix2vec(self._nside, pix, nest=self._nest)
        z += 1.0
        z *= 0.5
        z *= range
        z += self._min
        return z


class OpSimScan(Operator):
    """Operator which generates sky signal by scanning from a map.

    The signal to use should already be in a distributed pixel structure,
    and local pointing should already exist.

    Args:
        input_map (DistPixels or string):  Path to the map to load and
            sample.  If tag {detector} is encountered, it will be replaced
            with the actual detector name.
        pixels (str): the name of the cache object (<pixels>_<detector>)
            containing the pixel indices to use.
        weights (str): the name of the cache object (<weights>_<detector>)
            containing the pointing weights to use.
        nnz (int):  Number of non-zero weights per sample
        out (str): accumulate data to the cache with name <out>_<detector>.
            If the named cache objects do not exist, then they are created.
        mc (int):  Monte Carlo index used in synthezing file names

    """

    def __init__(
        self,
        distmap=None,
        input_map=None,
        pixels="pixels",
        weights="weights",
        nnz=3,
        out="scan",
        dets=None,
        mc=None,
    ):
        # Call the parent class constructor
        super().__init__()
        if input_map is None:
            raise RuntimeError("OpSimScan requires an input map")
        if distmap is not None:
            warnings.warn(
                "`distmap` is deprecated, please use `input_map`", DeprecationWarning
            )
            self._input_map = distmap
        else:
            self._input_map = input_map
        self._pixels = pixels
        self._weights = weights
        self._nnz = nnz
        self._out = out
        self._dets = dets
        self._mc = mc

    @function_timer
    def exec(self, data):
        """Create the timestreams by scanning from the map.

        This loops over all observations and detectors and uses the pointing
        matrix to project the distributed map into a timestream.

        Args:
            data (toast.Data): The distributed data.

        Returns:
            None

        """
        log = Logger.get()

        if isinstance(self._input_map, DistPixels):
            input_map = self._input_map
        elif "{detector}" not in self._input_map:
            fname = self._input_map.format(mc=self._mc)
            if data.comm is None or data.comm.world_rank == 0:
                log.info("Scanning {}".format(fname))
                if not os.path.isfile(fname):
                    raise RuntimeError("Input map not found: {}".format(fname))
            input_map = DistPixels(
                data, nnz=self._nnz, dtype=np.float32, pixels=self._pixels
            )
            input_map.read_healpix_fits(fname)
        else:
            input_map = None

        for obs in data.obs:
            tod = obs["tod"]

            if self._dets is None:
                dets = tod.local_dets
            else:
                dets = self._dets

            for det in dets:
                if input_map is None:
                    if MPI is None:
                        comm = None
                    else:
                        comm = MPI.COMM_SELF
                    filename = self._input_map.format(detector=det, mc=self._mc)
                    if not os.path.isfile(filename):
                        raise RuntimeError("Input map not found: {}".format(filename))
                    detector_map = DistPixels(
                        data,
                        comm=comm,
                        nnz=self._nnz,
                        dtype=np.float32,
                        pixels=self._pixels,
                    )
                    detector_map.read_healpix_fits(filename)
                else:
                    detector_map = input_map

                # get the pixels and weights from the cache

                pixelsname = "{}_{}".format(self._pixels, det)
                weightsname = "{}_{}".format(self._weights, det)
                pixels = tod.cache.reference(pixelsname)
                weights = tod.cache.reference(weightsname)

                nsamp, nnz = weights.shape

                gt = GlobalTimers.get()
                gt.start("OpSimScan.exec.global_to_local")
                sm, lpix = detector_map.global_to_local(pixels)
                gt.stop("OpSimScan.exec.global_to_local")

                maptod = np.zeros(nsamp)
                maptype = np.dtype(detector_map.dtype)
                gt.start("OpSimScan.exec.scan_map")
                if maptype.char == "d":
                    scan_map_float64(
                        detector_map.npix_submap,
                        nnz,
                        sm.astype(np.int64),
                        lpix.astype(np.int64),
                        detector_map.flatdata,
                        weights.astype(np.float64).reshape(-1),
                        maptod,
                    )
                elif maptype.char == "f":
                    scan_map_float32(
                        detector_map.npix_submap,
                        nnz,
                        sm.astype(np.int64),
                        lpix.astype(np.int64),
                        detector_map.flatdata,
                        weights.astype(np.float64).reshape(-1),
                        maptod,
                    )
                else:
                    raise RuntimeError(
                        "Scanning from a map only supports float32 and float64 maps"
                    )
                gt.stop("OpSimScan.exec.scan_map")

                cachename = "{}_{}".format(self._out, det)
                if not tod.cache.exists(cachename):
                    tod.cache.create(cachename, np.float64, (nsamp,))
                ref = tod.cache.reference(cachename)
                ref[:] += maptod

                del ref
                del pixels
                del weights

        return
